import pickle
import unittest
import tornado
import uuid

from unittest.mock import patch
from datetime import date, timedelta

from tornado.testing import AsyncTestCase
from chtoolset import query as chquery
from freezegun import freeze_time

from tinybird.constants import CHCluster
from tinybird.token_scope import scopes
from tinybird.user import UserAccount, UserDoesNotExist, Users, User, PipeWithoutEndpoint, QueryNotAllowed, public, DatabaseNameCollisionError
from tinybird.redis_client import get_redis_client
from tinybird.plans import BillingPlans
from tinybird.default_tables import Replacement, c, FeaturedColumns

from .conftest import CH_ADDRESS
from .views.base_test import BaseTest, TBApiProxyAsync


class TestUserModelWithRedisMaster(BaseTest):

    def setUp(self):
        USER = f'test_redis_{uuid.uuid4().hex}@example.com'
        WORKSPACE = f'test_workspace_redis_{uuid.uuid4().hex}'

        self.redis_client = get_redis_client()
        self.user_account = UserAccount.register(USER, 'pass')
        self.user = User.register(WORKSPACE, self.user_account.id)
        super().setUp()

    def tearDown(self):
        User._delete(self.user.id)
        UserAccount._delete(self.user_account.id)
        super().tearDown()

    def _get_from_redis(self, uid):
        f = self.redis_client.get('users:' + uid)
        return pickle.loads(f)

    def test_user_should_be_in_redis(self):
        u = self._get_from_redis(self.user['id'])
        self.assertEqual(u['id'], self.user['id'])

        u = Users.get_by_id(self.user['id'])
        self.assertEqual(u['id'], self.user['id'])

    def test_user_by_name(self):
        u = Users.get_by_name(self.user['name'])
        self.assertEqual(u['id'], self.user['id'])

    def test_user_by_database(self):
        u = Users.get_by_database(self.user['database'])
        self.assertEqual(u['id'], self.user['id'])

    def test_non_existing_user(self):
        with self.assertRaises(UserDoesNotExist):
            Users.get_by_id('asdasd')
        # self.assertEqual(u['id'], self.user['id'])

    def test_should_update(self):
        # change
        Users.add_datasource(self.user, 'testing')
        # read
        u = Users.get_by_id(self.user['id'])
        ds = Users.get_datasources(u)
        self.assertEqual(len(ds), 1)


class TestUserReplaceTablesTemplateSetting(BaseTest):

    scenarios = [
        (
            'limit_threads',
            [
                {
                    'name': 'endpoint',
                    'sql': "% {{ max_threads(3) }} select 1 as a"
                }
            ],
            3
        ),
        (
            'limit_threads_other_node',
            [
                {
                    'name': 'first_node',
                    'sql': "% {{ max_threads(17) }} select 1 as a"
                },
                {
                    'name': 'endpoint',
                    'sql': "SELECT * FROM first_node"
                }
            ],
            17
        )
    ]

    def _prepare_test(self, pipe_name, nodes):
        u = Users.get_by_id(self.WORKSPACE_ID)

        token_name = f'{pipe_name} endpoint api'
        with User.transaction(u.id) as user:
            pipe = user.add_pipe(pipe_name, nodes=nodes)
            endpoint_node = next((n for n in pipe.pipeline.nodes if n.name == 'endpoint'), None) or pipe.pipeline.nodes[-1]
            pipe.endpoint = endpoint_node.id
            user.update_pipe(pipe)
            user.add_token(token_name, scopes.PIPES_READ, pipe.id)

        u = Users.get_by_id(self.WORKSPACE_ID)
        pipe = u.get_pipe(pipe_name)
        access_info = u.get_token_access_info(token_name)
        readable_resources = access_info.get_resources_for_scope(scopes.DATASOURCES_READ, scopes.PIPES_READ)
        return u, pipe, readable_resources

    def test_template_max_threads_setting(self):
        for pipe_name, nodes, expected_max_threads in self.scenarios:
            with self.subTest(pipe_name=pipe_name):
                u, pipe, readable_resources = self._prepare_test(pipe_name, nodes)

                template_execution_results = {}
                _ = u.replace_tables(f'SELECT * FROM {pipe_name}',
                                     readable_resources=readable_resources,
                                     pipe=pipe,
                                     template_execution_results=template_execution_results)

                self.assertEqual(template_execution_results.get('max_threads', None), expected_max_threads)

    @tornado.testing.gen_test
    async def test_template_max_threads_setting_executor(self):
        for pipe_name, nodes, expected_max_threads in self.scenarios:
            with self.subTest(pipe_name=pipe_name):
                u, pipe, readable_resources = self._prepare_test(pipe_name, nodes)

                template_execution_results = {}
                _ = await u.replace_tables_async(f'SELECT * FROM {pipe_name}',
                                                 readable_resources=readable_resources,
                                                 pipe=pipe,
                                                 template_execution_results=template_execution_results)

                self.assertEqual(template_execution_results.get('max_threads', None), expected_max_threads)


class TestUserReplaceTablesTemplateBackendHintSetting(BaseTest):

    scenarios = [
        (
            'backend_hint',
            [
                {
                    'name': 'endpoint',
                    'sql': "% {{ backend_hint('foo') }} select 1 as a"
                }
            ],
            'foo'
        ),
        (
            'backend_hint_non_str',
            [
                {
                    'name': 'endpoint',
                    'sql': "% {{ backend_hint(1) }} select 1 as a"
                }
            ],
            '1'
        ),
        (
            'backend_hint_other_node',
            [
                {
                    'name': 'first_node',
                    'sql': "% {{ backend_hint('bar') }} select 1 as a"
                },
                {
                    'name': 'endpoint',
                    'sql': "SELECT * FROM first_node"
                }
            ],
            'bar'
        )
    ]

    def _prepare_test(self, pipe_name, nodes):
        u = Users.get_by_id(self.WORKSPACE_ID)

        token_name = f'{pipe_name} endpoint api'
        with User.transaction(u.id) as user:
            pipe = user.add_pipe(pipe_name, nodes=nodes)
            endpoint_node = next((n for n in pipe.pipeline.nodes if n.name == 'endpoint'), None) or pipe.pipeline.nodes[-1]
            pipe.endpoint = endpoint_node.id
            user.update_pipe(pipe)
            user.add_token(token_name, scopes.PIPES_READ, pipe.id)

        u = Users.get_by_id(self.WORKSPACE_ID)
        pipe = u.get_pipe(pipe_name)
        access_info = u.get_token_access_info(token_name)
        readable_resources = access_info.get_resources_for_scope(scopes.DATASOURCES_READ, scopes.PIPES_READ)
        return u, pipe, readable_resources

    def test_template_backend_hint_setting(self):
        for pipe_name, nodes, expected_backend_hint in self.scenarios:
            with self.subTest(pipe_name=pipe_name):
                u, pipe, readable_resources = self._prepare_test(pipe_name, nodes)

                template_execution_results = {}
                _ = u.replace_tables(f'SELECT * FROM {pipe_name}',
                                     readable_resources=readable_resources,
                                     pipe=pipe,
                                     template_execution_results=template_execution_results)

                self.assertEqual(template_execution_results.get('backend_hint', None), expected_backend_hint)

    @tornado.testing.gen_test
    async def test_template_backend_hint_setting_executor(self):
        for pipe_name, nodes, expected_backend_hint in self.scenarios:
            with self.subTest(pipe_name=pipe_name):
                u, pipe, readable_resources = self._prepare_test(pipe_name, nodes)

                template_execution_results = {}
                _ = await u.replace_tables_async(f'SELECT * FROM {pipe_name}',
                                                 readable_resources=readable_resources,
                                                 pipe=pipe,
                                                 template_execution_results=template_execution_results)

                self.assertEqual(template_execution_results.get('backend_hint', None), expected_backend_hint)


class TestUserReplaceTablesBasic(BaseTest):

    def setUp(self):
        super(TestUserReplaceTablesBasic, self).setUp()
        self.tb_api_proxy_async = TBApiProxyAsync(self)

    def test_migrated(self):
        user_name = f'test_replace_tables_{uuid.uuid4().hex}'
        user_account = UserAccount.register(f'{user_name}@example.com', 'pass')
        u = User.register(user_name, user_account.id)

        ds = Users.add_datasource(u, 'foo')

        sql_from_name = Users.replace_tables(u, "select * from foo")
        sql_from_db_id = chquery.format(Users.replace_tables(u, f"select * from {u['database']}.{ds.id} AS foo"))
        self.assertEqual(sql_from_name, sql_from_db_id)

        pipe = Users.add_pipe(u, 'test_pipe', 'select * from foo')
        with self.assertRaisesRegex(PipeWithoutEndpoint, r'does not have an endpoint yet'):
            Users.replace_tables(u, f"select * from {pipe.name}")

        pipe.endpoint = pipe.pipeline.nodes[0].id
        Users.update_pipe(u, pipe)

        expected_query_with_name = chquery.format(f"select * from (SELECT * FROM {u['database']}.{ds.id} AS foo) as {pipe.name}")
        expected_query_with_id = chquery.format(f"select * from (SELECT * FROM {u['database']}.{ds.id} AS foo) as {pipe.id}")
        self.assertEqual(Users.replace_tables(u, f"select * from {pipe.name}"), expected_query_with_name)
        self.assertEqual(Users.replace_tables(u, f"select * from {pipe.id}"), expected_query_with_id)

        with self.assertRaisesRegex(QueryNotAllowed, rf"Resource '{pipe.endpoint}' not found"):
            Users.replace_tables(u, f"select * from {pipe.endpoint}")

        self.assertEqual(Users.replace_tables(u, f"select * from {pipe.id}", pipe=pipe), expected_query_with_id)

        with self.assertRaisesRegex(QueryNotAllowed, r"Resource 'system.block_log' not found"):
            Users.replace_tables(u, "select * from system.block_log")

        with self.assertRaisesRegex(QueryNotAllowed, r"Resource 'block_log' not found"):
            Users.replace_tables(u, "select * from block_log")

        q = Users.replace_tables(u, "select * from tinybird.block_log")
        self.assertIsNotNone(q)

    @tornado.testing.gen_test
    async def test_shared_internal_ds(self):
        ws_public = public.get_public_user()
        public_user_account = UserAccount.get_by_email(public.get_public_email())
        user_name = f'test_replace_shared_internal_tables_{uuid.uuid4().hex}'
        user_account = UserAccount.register(f'{user_name}@example.com', 'pass')
        workspace = User.register(user_name, user_account.id, CHCluster('tinybird', User.default_database_server))
        token_user_account = UserAccount.get_token_for_scope(user_account, scopes.AUTH)

        # before sharing replacements include filters (i.e billable)
        q = Users.replace_tables(workspace, "select * from tinybird.pipe_stats")

        self.assertIn('AND (billable = 1)', q)
        token = UserAccount.get_token_for_scope(public_user_account, scopes.AUTH)
        await self.tb_api_proxy_async.invite_user_to_workspace(
            token=token,
            workspace_id=ws_public.id,
            user_to_invite_email=user_account.email
        )
        await self.tb_api_proxy_async.make_user_workspace_admin(token=token, workspace_id=ws_public.id, user_email=user_account.email)
        await self.tb_api_proxy_async.share_datasource_with_another_workspace(
            token=token_user_account,
            datasource_id=ws_public.get_datasource('pipe_stats').id,
            origin_workspace_id=ws_public.id,
            destination_workspace_id=workspace.id,
            expect_notification=False
        )

        workspace = User.get_by_id(workspace.id)
        self.assertEqual(len(workspace.get_datasources()), 1)
        self.assertEqual(workspace.get_datasources()[0].original_ds_name, 'pipe_stats')

        # after sharing replacements doesn't include filters (i.e billable)
        q = Users.replace_tables(workspace, "select * from tinybird.pipe_stats")
        self.assertNotIn('AND (billable = 1)', q)

    @tornado.testing.gen_test
    async def test_feature_columns_replacements(self):
        # TODO: make this test less data dependant
        user_name = f'test_replace_feature_columns_tables_{uuid.uuid4().hex}'
        user_account = UserAccount.register(f'{user_name}@example.com', 'pass')
        workspace = User.register(user_name, user_account.id, CHCluster('tinybird', User.default_database_server))

        mock_replacements = [Replacement(
            namespace='tinybird',
            name='datasources_ops_log',
            resource='datasources_ops_log',
            columns=c('timestamp', 'event_type', 'datasource_id', 'datasource_name', 'result', 'elapsed_time',
                      'error', 'Options.Names', 'Options.Values', 'request_id', 'import_id', 'job_id', 'rows',
                      'rows_quarantine', 'blocks_ids'),
            filter_by_user=True,
            include_read_only_datasources=True,
            featured_columns=FeaturedColumns(feature_flag="ANY_FEATURED_FLAG",
                                             columns=c('timestamp', 'event_type', 'datasource_id', 'datasource_name',
                                                       'result', 'elapsed_time',
                                                       'error', 'Options.Names', 'Options.Values', 'request_id',
                                                       'import_id', 'job_id', 'rows', 'rows_quarantine', 'blocks_ids',
                                                       'operation_id', 'read_rows', 'read_bytes', 'written_rows',
                                                       'written_bytes', 'written_rows_quarantine',
                                                       'written_bytes_quarantine',
                                                       'pipe_id', 'pipe_name')),
        )]

        with patch('tinybird.user.REPLACEMENTS', mock_replacements):
            with patch('tinybird.user.FeatureFlagsWorkspaceService.feature_for_id', return_value=False):
                q = Users.replace_tables(workspace, "select * from tinybird.datasources_ops_log")
                self.assertNotIn('operation_id', q)

        with patch('tinybird.user.REPLACEMENTS', mock_replacements):
            with patch('tinybird.user.FeatureFlagsWorkspaceService.feature_for_id', return_value=True):
                q = Users.replace_tables(workspace, "select * from tinybird.datasources_ops_log")
                self.assertIn('operation_id', q)

    @tornado.testing.gen_test
    async def test_replace_filtering_by_database(self):
        user_name = f'test_replace_filter_by_database_{uuid.uuid4().hex}'
        user_account = UserAccount.register(f'{user_name}@example.com', 'pass')
        workspace = User.register(user_name, user_account.id, CHCluster('tinybird', User.default_database_server))
        q = Users.replace_tables(workspace, "select * from tinybird.bi_stats_rt")
        self.assertIn(f"WHERE database = '{workspace.database}'", q)


class TestUserSetNameFromNotNormalizedEmail(unittest.TestCase):

    @staticmethod
    def _user_workspace_tear_down(email, name):
        try:
            u = UserAccount.get_by_email(email)
            UserAccount._delete(u.id)
        except Exception:
            pass

        try:
            u = User.get_by_name(name)
            User._delete(u.id)
        except Exception:
            pass

    def test_happy_path(self):
        name = f'test.email_{uuid.uuid4().hex}'
        name_replaced = name.replace('.', '_')
        email = f'{name}@example.com'
        self._user_workspace_tear_down(email, name_replaced)

        user_account = UserAccount.register(email, 'pass')
        workspace = User.register(name=name, normalize_name_and_try_different_on_collision=True, admin=user_account.id)

        refresed_workspace = User.get_by_id(workspace.id)
        self.assertEqual(refresed_workspace.name, name_replaced)

    @patch('tinybird.user.User._get_string_with_random_numbers', side_effect=lambda value: '9' * value)
    def test_case_where_the_normalized_email_is_already_taken(self, _get_string_with_random_numbers_mock):
        first_name = f'test.email_{uuid.uuid4().hex}'
        email = f'{first_name}@example.com'
        second_name = first_name

        self._user_workspace_tear_down(email, first_name)

        user_account = UserAccount.register(email, 'pass')
        first_workspace = User.register(name=first_name, normalize_name_and_try_different_on_collision=True, admin=user_account.id)
        second_workspace = User.register(name=second_name, normalize_name_and_try_different_on_collision=True, admin=user_account.id)

        refresed_first_workspace = User.get_by_id(first_workspace.id)
        self.assertEqual(refresed_first_workspace.name, first_name.replace('.', '_'))

        refresed_second_workspace = User.get_by_id(second_workspace.id)
        self.assertEqual(refresed_second_workspace.name, f'{first_name}_99'.replace('.', '_'))


class TestWorkspaceDatabaseColission(unittest.TestCase):

    def test_create_database_collision(self):
        with patch('tinybird.user.User.get_by_database', return_value=None):
            with self.assertRaises(DatabaseNameCollisionError):
                User.generate_uid_and_database_name_and_try_different_on_collision()


class TestWorkspaceDelete(unittest.TestCase):
    def setUp(self):
        self.email = f'test_workspace_deletion_{uuid.uuid4().hex}@example.com'
        self.workspace_name = f'test_workspace_deletion_{uuid.uuid4().hex}'

        self.redis_client = get_redis_client()
        self.user_account = UserAccount.register(self.email, 'pass')
        self.workspace = User.register(self.workspace_name, self.user_account.id)

    def tearDown(self):
        User._delete(self.workspace.id)
        UserAccount._delete(self.user_account.id)

    def test_delete_happy_path(self):
        Users.delete(self.workspace)
        refreshed_workspace = User.get_by_id(self.workspace.id)

        self.assertTrue(refreshed_workspace.deleted)
        self.assertFalse(refreshed_workspace.confirmed_account)
        self.assertIn('deleted', refreshed_workspace.name)

    def test_once_a_ws_is_deleted_its_name_can_be_reused(self):
        name_before_deletion = self.workspace.name
        Users.delete(self.workspace)
        refreshed_workspace = User.get_by_id(self.workspace.id)
        self.assertIn('deleted', refreshed_workspace.name)

        the_new_workspace_created = User.register(name_before_deletion, self.user_account.id)
        self.assertIsNotNone(the_new_workspace_created)


class TestUserGetClusters(AsyncTestCase):

    def setUp(self):
        super(TestUserGetClusters, self).setUp()

        self.workspaces_to_delete = []
        self.users_to_delete = []

        self.main_clickhouse_url = CH_ADDRESS

    def tearDown(self):
        for ws in self.workspaces_to_delete:
            User._delete(ws.id)
        for ua in self.users_to_delete:
            UserAccount._delete(ua.id)
        super(TestUserGetClusters, self).tearDown()

    @tornado.testing.gen_test
    async def test_get_cluster_returns_default_for_non_configured_emails(self):
        user = UserAccount.register(f'test{uuid.uuid4().hex}@developeruser.co')
        self.users_to_delete.append(user)

        returned_cluster = await user.get_cluster(self.main_clickhouse_url)

        self.assertEqual(CHCluster(name=User.default_cluster, server_url=User.default_database_server), returned_cluster)

    @tornado.testing.gen_test
    @patch('tinybird.user.OrganizationEmailsCluster', {'developeruser.co': CHCluster('tinybird', '127.0.0.1')})
    async def test_get_cluster_returns_configuration_for_email(self):
        user = UserAccount.register(f'test{uuid.uuid4().hex}@developeruser.co')
        self.users_to_delete.append(user)

        returned_cluster = await user.get_cluster(self.main_clickhouse_url)
        self.assertEqual('127.0.0.1', returned_cluster.server_url)
        self.assertEqual('tinybird', returned_cluster.name)

    @tornado.testing.gen_test
    @patch('tinybird.user.OrganizationEmailsCluster', {'developeruser.co': CHCluster('non_existing_cluster', '127.0.0.1')})
    async def test_get_cluster_returns_default_cluster_if_configured_cluster_doesnt_exists(self):
        user = UserAccount.register(f'test{uuid.uuid4().hex}@developeruser.co')
        self.users_to_delete.append(user)

        self.assertEqual(CHCluster(name=User.default_cluster, server_url=User.default_database_server), await user.get_cluster(self.main_clickhouse_url))

    @tornado.testing.gen_test
    @patch('tinybird.user.OrganizationEmailsCluster', {'developeruser.co': CHCluster('tinybird', '127.0.0.1')})
    async def test_get_cluster_returns_cluster_used_in_the_first_workspace_created_by_the_user_even_when_email_config_is_present__workspace_with_cluster(self):
        user = UserAccount.register(f'test{uuid.uuid4().hex}@developeruser.co')
        self.users_to_delete.append(user)

        new_workspace_1 = User.register(name=f"test{uuid.uuid4().hex}", admin=user.id)
        self.workspaces_to_delete.append(new_workspace_1)
        new_workspace_2 = User.register(name=f"test{uuid.uuid4().hex}", admin=user.id)

        self.workspaces_to_delete.append(new_workspace_2)

        with User.transaction(new_workspace_1.id) as w:
            w.clusters = ['cluster_from_oldest_ws']
            w.database_server = '1.1.1.1'

        with User.transaction(new_workspace_2.id) as w:
            w.clusters = ['cluster_from_newest_ws']
            w.database_server = '1.1.1.2'

        returned_cluster = await user.get_cluster(self.main_clickhouse_url)
        self.assertEqual('1.1.1.1', returned_cluster.server_url)
        self.assertEqual('cluster_from_oldest_ws', returned_cluster.name)


class TestNotifyBuildPlanLimits(AsyncTestCase):

    def setUp(self):
        super(TestNotifyBuildPlanLimits, self).setUp()
        self.workspaces_to_delete = []
        self.users_to_delete = []
        user_account = UserAccount.register(f"test_notify_build_plan{uuid.uuid4().hex}@example.com", 'pass')
        self.users_to_delete.append(user_account)
        self.first_workspace = User.register(
            name=f"test_workspace_notify_build_plan_{uuid.uuid4().hex}",
            admin=user_account.id)
        self.workspaces_to_delete.append(self.first_workspace)

    def tearDown(self):
        for ws in self.workspaces_to_delete:
            User._delete(ws.id)
        for ua in self.users_to_delete:
            UserAccount._delete(ua.id)
        super(TestNotifyBuildPlanLimits, self).tearDown()

    @tornado.testing.gen_test
    @freeze_time('2022-03-30')
    @patch('tinybird.user.Users._send_notification_on_build_plan_limits')
    async def test_notify_build_plan_limits_just_upgraded(self, mock_send_notification):
        self.first_workspace.plan = BillingPlans.PRO
        self.first_workspace.save()
        await Users.notify_build_plan_limits(self.first_workspace.id, 1000, 10, 0.07, 0.34, False, 800, 7)
        mock_send_notification.assert_not_called()
        self.first_workspace = Users.get_by_id(self.first_workspace.id)
        self.assertIsNone(self.first_workspace.billing_details.get('notifications'))

    @tornado.testing.gen_test
    @freeze_time('2022-03-30')
    @patch('tinybird.user.Users._send_notification_on_build_plan_limits')
    async def test_notify_build_plan_limits_reaching_empty_notifications(self, mock_send_notification):
        await Users.notify_build_plan_limits(self.first_workspace.id, 1000, 10, 0.07, 0.34, False, 800, 7)
        mock_send_notification.assert_called_once_with(self.first_workspace, 1000, 10, 0.07, 0.34,
                                                       False, 800, 7)
        self.first_workspace = Users.get_by_id(self.first_workspace.id)
        self.assertEqual(self.first_workspace.billing_details['notifications'], {'plan_reaching': date.today()})

    @tornado.testing.gen_test
    @freeze_time('2022-03-30')
    @patch('tinybird.user.Users._send_notification_on_build_plan_limits')
    async def test_notify_build_plan_limits_reaching_already_reaching_sent(self, mock_send_notification):
        self.first_workspace.billing_details = {'notifications': {'plan_reaching': date.today()}}
        self.first_workspace.save()
        await Users.notify_build_plan_limits(self.first_workspace.id, 1000, 10, 0.07, 0.34, False, 800, 7)
        mock_send_notification.assert_not_called()
        self.first_workspace = Users.get_by_id(self.first_workspace.id)
        self.assertEqual(self.first_workspace.billing_details['notifications'], {'plan_reaching': date.today()})

    @tornado.testing.gen_test
    @patch('tinybird.user.Users._send_notification_on_build_plan_limits')
    async def test_notify_build_plan_limits_reaching_already_reaching_sent_yesterday(self, mock_send_notification):
        self.first_workspace.billing_details = {'notifications': {'plan_reaching': date.today() - timedelta(days=1)}}
        self.first_workspace.save()
        await Users.notify_build_plan_limits(self.first_workspace.id, 1000, 10, 0.07, 0.34, False, 800, 7)
        mock_send_notification.assert_called_once_with(self.first_workspace, 1000, 10, 0.07, 0.34,
                                                       False, 800, 7)
        self.first_workspace = Users.get_by_id(self.first_workspace.id)
        self.assertEqual(self.first_workspace.billing_details['notifications'], {'plan_reaching': date.today()})

    @tornado.testing.gen_test
    @freeze_time('2022-03-30')
    @patch('tinybird.user.Users._send_notification_on_build_plan_limits')
    async def test_notify_build_plan_limits_reaching_already_reaching_sent_yesterday_only_requests(self, mock_send_notification):
        self.first_workspace.billing_details = {'notifications': {'plan_reaching': date.today() - timedelta(days=1)}}
        self.first_workspace.save()
        await Users.notify_build_plan_limits(self.first_workspace.id, 1000, 10, 0.07, 0.34, False, 800, None)
        mock_send_notification.assert_called_once_with(self.first_workspace, 1000, 10, 0.07, 0.34,
                                                       False, 800, None)
        self.first_workspace = Users.get_by_id(self.first_workspace.id)
        self.assertEqual(self.first_workspace.billing_details['notifications'], {'plan_reaching': date.today()})

    @tornado.testing.gen_test
    @freeze_time('2022-03-30')
    @patch('tinybird.user.Users._send_notification_on_build_plan_limits')
    async def test_notify_build_plan_limits_reaching_already_reaching_sent_yesterday_only_storage(self, mock_send_notification):
        self.first_workspace.billing_details = {'notifications': {'plan_reaching': date.today() - timedelta(days=1)}}
        self.first_workspace.save()
        await Users.notify_build_plan_limits(self.first_workspace.id, 1000, 10, 0.07, 0.34, False, None, 7)
        mock_send_notification.assert_not_called()
        self.first_workspace = Users.get_by_id(self.first_workspace.id)
        self.assertEqual(self.first_workspace.billing_details['notifications'], {'plan_reaching': date.today() - timedelta(days=1)})

    @tornado.testing.gen_test
    @freeze_time('2022-03-30')
    @patch('tinybird.user.Users._send_notification_on_build_plan_limits')
    async def test_notify_build_plan_limits_reaching_already_reaching_sent_prev_month_only_storage(self, mock_send_notification):
        self.first_workspace.billing_details = {'notifications': {'plan_reaching': date.today() - timedelta(days=40)}}
        self.first_workspace.save()
        await Users.notify_build_plan_limits(self.first_workspace.id, 1000, 10, 0.07, 0.34, False, None, 7)
        mock_send_notification.assert_called_once_with(self.first_workspace, 1000, 10, 0.07, 0.34,
                                                       False, None, 7)
        self.first_workspace = Users.get_by_id(self.first_workspace.id)
        self.assertEqual(self.first_workspace.billing_details['notifications'], {'plan_reaching': date.today()})

    @tornado.testing.gen_test
    @freeze_time('2022-03-30')
    @patch('tinybird.user.Users._send_notification_on_build_plan_limits')
    async def test_notify_build_plan_limits_exceeded_already_reaching_sent(self, mock_send_notification):
        self.first_workspace.billing_details = {'notifications': {'plan_reaching': date.today()}}
        self.first_workspace.save()
        await Users.notify_build_plan_limits(self.first_workspace.id, 1000, 10, 0.07, 0.34, True, 800, 7)
        mock_send_notification.assert_called_once_with(self.first_workspace, 1000, 10, 0.07, 0.34,
                                                       True, 800, 7)
        self.first_workspace = Users.get_by_id(self.first_workspace.id)
        self.assertEqual(self.first_workspace.billing_details['notifications'], {'plan_reaching': date.today(),
                                                                                 'plan_exceeded': date.today()})

    @tornado.testing.gen_test
    @freeze_time('2022-03-30')
    @patch('tinybird.user.Users._send_notification_on_build_plan_limits')
    async def test_notify_build_plan_limits_exceeded_already_exceeded_sent(self, mock_send_notification):
        self.first_workspace.billing_details = {'notifications': {'plan_reaching': date.today(),
                                                                  'plan_exceeded': date.today()}
                                                }
        self.first_workspace.save()
        await Users.notify_build_plan_limits(self.first_workspace.id, 1000, 10, 0.07, 0.34, True, 800, 7)
        mock_send_notification.assert_not_called()
        self.first_workspace = Users.get_by_id(self.first_workspace.id)
        self.assertEqual(self.first_workspace.billing_details['notifications'], {'plan_reaching': date.today(),
                                                                                 'plan_exceeded': date.today()})

    @tornado.testing.gen_test
    @freeze_time('2022-03-30')
    @patch('tinybird.user.Users._send_notification_on_build_plan_limits')
    async def test_notify_build_plan_limits_exceeded_already_exceeded_sent_yesterday(self, mock_send_notification):
        self.first_workspace.billing_details = {'notifications': {'plan_reaching': date.today() - timedelta(days=1),
                                                                  'plan_exceeded': date.today() - timedelta(days=1)}
                                                }
        self.first_workspace.save()
        await Users.notify_build_plan_limits(self.first_workspace.id, 1000, 10, 0.07, 0.34, True, 800, 7)
        mock_send_notification.assert_called_once_with(self.first_workspace, 1000, 10, 0.07, 0.34,
                                                       True, 800, 7)
        self.first_workspace = Users.get_by_id(self.first_workspace.id)
        self.assertEqual(self.first_workspace.billing_details['notifications'], {'plan_reaching': date.today() - timedelta(days=1),
                                                                                 'plan_exceeded': date.today()})

    @tornado.testing.gen_test
    @freeze_time('2022-03-30')
    @patch('tinybird.user.Users._send_notification_on_build_plan_limits')
    async def test_notify_build_plan_limits_exceeded_already_exceeded_sent_yesterday_only_requests(self, mock_send_notification):
        self.first_workspace.billing_details = {'notifications': {'plan_reaching': date.today() - timedelta(days=1),
                                                                  'plan_exceeded': date.today() - timedelta(days=1)}
                                                }
        self.first_workspace.save()
        await Users.notify_build_plan_limits(self.first_workspace.id, 1000, 10, 0.07, 0.34, True, 800, None)
        mock_send_notification.assert_called_once_with(self.first_workspace, 1000, 10, 0.07, 0.34,
                                                       True, 800, None)
        self.first_workspace = Users.get_by_id(self.first_workspace.id)
        self.assertEqual(self.first_workspace.billing_details['notifications'], {'plan_reaching': date.today() - timedelta(days=1),
                                                                                 'plan_exceeded': date.today()})

    @tornado.testing.gen_test
    @freeze_time('2022-03-30')
    @patch('tinybird.user.Users._send_notification_on_build_plan_limits')
    async def test_notify_build_plan_limits_exceeded_already_exceeded_sent_yesterday_only_storage(self, mock_send_notification):
        self.first_workspace.billing_details = {'notifications': {'plan_reaching': date.today() - timedelta(days=1),
                                                                  'plan_exceeded': date.today() - timedelta(days=1)}
                                                }
        self.first_workspace.save()
        await Users.notify_build_plan_limits(self.first_workspace.id, 1000, 10, 0.07, 0.34, True, None, 11)
        mock_send_notification.assert_not_called()
        self.first_workspace = Users.get_by_id(self.first_workspace.id)
        self.assertEqual(self.first_workspace.billing_details['notifications'], {'plan_reaching': date.today() - timedelta(days=1),
                                                                                 'plan_exceeded': date.today() - timedelta(days=1)})

    @tornado.testing.gen_test
    @freeze_time('2022-03-30')
    @patch('tinybird.user.Users._send_notification_on_build_plan_limits')
    async def test_notify_build_plan_limits_exceeded_already_exceeded_sent_prev_month_only_storage(self, mock_send_notification):
        self.first_workspace.billing_details = {'notifications': {'plan_reaching': date.today() - timedelta(days=1),
                                                                  'plan_exceeded': date.today() - timedelta(days=40)}
                                                }
        self.first_workspace.save()
        await Users.notify_build_plan_limits(self.first_workspace.id, 1000, 10, 0.07, 0.34, True, None, 11)
        mock_send_notification.assert_called_once_with(self.first_workspace, 1000, 10, 0.07, 0.34,
                                                       True, None, 11)
        self.first_workspace = Users.get_by_id(self.first_workspace.id)
        self.assertEqual(self.first_workspace.billing_details['notifications'], {'plan_reaching': date.today() - timedelta(days=1),
                                                                                 'plan_exceeded': date.today()})



from uuid import uuid4
from tornado.testing import AsyncTestCase
import tornado
from tests_e2e.aux import poll_async
from tinybird.async_redis import async_redis_writer
from tinybird.redis_client import get_redis_args
from tinybird.distributed import WorkingGroup


class TestDistributed(AsyncTestCase):
    @tornado.testing.gen_test
    async def test_workinggroup(self):
        redis_args = get_redis_args()
        async_redis_writer.init({
            'redis_host': redis_args['host'],
            'redis_db': redis_args['db']})
        workinggroup_id = f"test_wg_{uuid4()}"
        wg1 = await WorkingGroup(workinggroup_id, 'worker1', ttl=1, keepalive_interval=0.1).init()
        assert wg1.score_index('k0') == 0
        wg2 = await WorkingGroup(workinggroup_id, 'worker2', ttl=1, keepalive_interval=0.1).init()

        async def f():
            assert {wg1.score_index('k0'), wg2.score_index('k0')} == {0, 1}

        await poll_async(f)

        if wg1.score_index('k0') == 1:
            wg1, wg2 = wg2, wg1
        # wg1 has lower priority, closing wg2 to check priority changes (as wg1 disappears)
        await wg2.exit()

        async def f2():
            assert wg1.score_index('k0') == 0
            assert len(wg1._workers) == 1

        await poll_async(f2)

        await wg1.exit()
        await async_redis_writer.close()



from multiprocessing import Manager
import unittest
from time import time
from typing import Set, Callable
import uuid

from tinybird.ch import ch_create_materialized_view_sync, HTTPClient
from tinybird.ch_utils.engine import engine_local_to_replicated
from tinybird.constants import CHCluster
from tinybird.job import Job, JobStatus, new_populate_job, PopulateJob, JobCancelledException, JobExecutor
from tinybird.user import User, UserAccount
from tinybird.redis_client import get_redis_client, get_redis_args

from .utils import CH_ADDRESS, get_finalised_job_async, exec_sql, wait_until_job_is_in_expected_status_async

# Global Manager to handle locks across different instances of the same Job,
# potentially from different threads. This is used instead of `threading.Event`
# because it can be serialized/deserialized into/from Redis without problems.
# This way, both the `FakeTestJob` and the deserialized jobs collected using
# `Job.get_by_id` share the same lock and can be synchronized.
global_lock_manager = Manager()


class ExitJobWithoutMarkingItAsDone(Exception):
    pass


class FakeTestJob(Job):

    def __init__(self, user, cancellable_status: Set[JobStatus] = None, block_job_before_executing_it=False):
        self.user = user  # workspace FIXME?
        self.database_server = user['database_server']
        self.database = user['database']
        self.internal_result = None

        self._cancellable_status = cancellable_status

        # Set of events used to block the job in different statuses so it's easier to test things
        self.before_execution_event = global_lock_manager.Event()
        self.block_job_before_executing_it = block_job_before_executing_it
        self.job_already_blocked_before_execution = global_lock_manager.Event()

        self.exit_event = global_lock_manager.Event()

        # Event to wait until a JOB is finished, either as DONE, ERROR or CANCELLED
        self.mark_as_finalised_event = global_lock_manager.Event()
        self.mark_as_working_event = global_lock_manager.Event()

        self.exit_from_block_before_execution = False
        self.do_not_submit = False

        Job.__init__(self, "test", user)

    def mark_as_done(self, result, stats):
        super().mark_as_done(result, stats)
        self.mark_as_finalised_event.set()

    def mark_as_error(self, result):
        super().mark_as_error(result)
        self.mark_as_finalised_event.set()

    def mark_as_cancelled(self):
        super().mark_as_cancelled()
        self.mark_as_finalised_event.set()

    def mark_as_working(self):
        super().mark_as_working()
        self.mark_as_working_event.set()

    def wait_until_finalised(self):
        self.mark_as_finalised_event.wait()
        return Job.get_by_id(self.id)

    def wait_until_working(self):
        self.mark_as_working_event.wait()
        return Job.get_by_id(self.id)

    def to_json(self, u=None, debug=None):
        d = {
            "id": self.id,
            "status": self.status
        }
        return d

    def finish(self, result=None):
        self.internal_result = result
        self.save()
        self.exit_event.set()

    def run(self):
        def done(f):
            try:
                result = f.result()
            except JobCancelledException:
                self.mark_as_cancelled()
            except ExitJobWithoutMarkingItAsDone:
                return
            except Exception as e:
                try:
                    self.mark_as_error({'error': str(e)})
                except Exception:
                    pass
            else:
                self.mark_as_done(result, None)

        def function():
            def _wrapper(job):
                return FakeTestJob.run_until_event(job)
            return _wrapper

        if not self.do_not_submit:
            future = self.job_executor.submit(function_to_execute=function(), job=self)
            future.add_done_callback(done)
        return self

    @staticmethod
    def run_until_event(j):
        j = Job.get_by_id(j.id)
        if j.block_job_before_executing_it:
            if j.exit_from_block_before_execution:
                raise ExitJobWithoutMarkingItAsDone()
            j.job_already_blocked_before_execution.set()
            j.before_execution_event.wait()

        j.mark_as_working()

        j.exit_event.wait()
        j = Job.get_by_id(j.id)
        if j.status == JobStatus.CANCELLING:
            raise JobCancelledException()

        if isinstance(j.internal_result, Exception):
            raise j.internal_result
        return j.internal_result

    @property
    def is_cancellable(self) -> bool:
        if self._cancellable_status is None:
            return super(FakeTestJob, self).is_cancellable
        return self.status in self._cancellable_status

    def notify_cancelled_event(self):
        self.exit_event.set()

    def set_do_not_submit(self):
        self.do_not_submit = True
        self.save()


class FailOnRunFakeJob(FakeTestJob):
    def run(self):
        raise RuntimeError('Failing before going into executor')


class BaseTestJobs(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.job_executor = JobExecutor(redis_client=get_redis_client(), redis_kwargs=get_redis_args(), consumer=True, import_workers=1, query_workers=1)

    def tearDown(self):
        self.job_executor.join(wait=True)
        self.job_executor._clean()

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        WORKSPACE = f'test_jobs{str(int(time()))}_{uuid.uuid4().hex}'
        USER = f'{WORKSPACE}@example.com'
        cls.user_account = UserAccount.register(USER, 'pass')

        cls.cluster_clause = "ON CLUSTER tinybird"
        cls.cluster = CHCluster(name='tinybird', server_url=CH_ADDRESS)

        cls.user = User.register(
            name=WORKSPACE,
            admin=cls.user_account.id,
            cluster=cls.cluster
        )

        client = HTTPClient(cls.user.database_server, database=None)
        client.query_sync(f"CREATE DATABASE IF NOT EXISTS {cls.user.database} {cls.cluster_clause}", read_only=False)

    @classmethod
    def tearDownClass(cls):
        database = cls.user.database
        client = HTTPClient(cls.user.database_server, database=cls.user.database)
        User._delete(cls.user.id)
        UserAccount._delete(cls.user_account.id)
        client.query_sync(f"DROP DATABASE IF EXISTS `{database}` {cls.cluster_clause}", read_only=False)
        super().tearDownClass()


class TestJobsCorruption(BaseTestJobs):

    def setUp(self):
        super().setUp()
        self.job_consumer = self.job_executor.start_consumer()

    def tearDown(self):
        self.job_consumer.terminate()
        self.job_consumer.join()
        super().tearDown()

    def test_happy_case(self):
        j = FakeTestJob(self.user)
        j.save()

        self.assertEqual(j.status, 'waiting')
        self.job_executor.put_job(j)

        j.finish({'task': 'ok'})

        job = j.wait_until_finalised()
        self.assertEqual(job.status, 'done')
        self.assertEqual(job.result['task'], 'ok')

    def test_happy_case_exception(self):
        j = FakeTestJob(self.user)
        j.save()

        self.assertEqual(j.status, 'waiting')
        self.job_executor.put_job(j)

        j.finish(ValueError('broken job'))

        job = j.wait_until_finalised()
        self.assertEqual(job.status, 'error')
        self.assertEqual(job.result['error'], 'broken job')

    async def test_after_populate_failure(self):
        j = await new_populate_job(self.job_executor, self.user, 'will_crash', 'select 1', 'will_crash', 'id', 'name', unlink_on_populate_error=False)

        job = await get_finalised_job_async(j.id)
        self.assertEqual(job.status, 'error')
        self.assertEqual(job.datasource, None)
        self.assertEqual(job.result['error'], "Could not find target table 'will_crash'")

        another_job = FakeTestJob(self.user)
        another_job.save()
        self.job_executor.put_job(another_job)

        another_job.finish({'task': 'ok'})
        job = another_job.wait_until_finalised()
        self.assertEqual(job.status, 'done')
        self.assertEqual(job.result['task'], 'ok')

    def test_job_failure_before_executor(self):
        j = FailOnRunFakeJob(self.user)
        j.save()

        self.assertEqual(j.status, 'waiting')
        self.job_executor.put_job(j)

        job = j.wait_until_finalised()
        self.assertEqual(job.status, 'error')
        self.assertEqual(job.result['error'], 'Job failed to finish for an unknown reason. Try again.')


class TestJobsPopulatePartitions(BaseTestJobs):

    async def _test_by_partition_populate(self, partition_key_source='', partition_key_target=''):
        extra_params = {'output_format_json_quote_64bit_integers': 0}
        cluster_clause = f"ON CLUSTER {self.user.cluster}" if self.user.cluster else ""

        id = uuid.uuid4().hex
        table_source = f'test_by_partition_populate_source_{id}'
        table_target = f'test_by_partition_populate_target_{id}'
        view_node_id = f'test_by_partition_populate_mv_{id}'

        self.user.drop_datasource(table_source)
        self.user.drop_datasource(table_target)
        table_source_ds = self.user.add_datasource(table_source)
        table_target_ds = self.user.add_datasource(table_target)

        table_source_id = table_source_ds.id
        table_target_id = table_target_ds.id

        tables = [table_source_id, table_target_id, view_node_id]
        for t in tables:
            exec_sql(self.user.database, f"DROP TABLE IF EXISTS {t} {cluster_clause}")

        engine = f"MergeTree() {partition_key_source} ORDER BY tuple()"
        if self.user.cluster:
            engine = engine_local_to_replicated(engine, self.user['database'], table_source_id)
        exec_sql(self.user.database, f"CREATE TABLE {table_source_id} {cluster_clause} (ts DateTime('UTC'), n Int32) Engine = {engine}")

        engine = f"SummingMergeTree() {partition_key_target} ORDER BY d"
        if self.user.cluster:
            engine = engine_local_to_replicated(engine, self.user['database'], table_target_id)
        exec_sql(self.user.database, f"CREATE TABLE {table_target_id} {cluster_clause} (d Date, n Int32) Engine = {engine}")

        # Only odd hours of odd days of the month:
        # - 1st of January has 12 hours with data
        # - 4th of February has 0 hours with data
        exec_sql(self.user.database, f"INSERT INTO {table_source_id} SELECT toDateTime('2021-01-01 00:00:00', 'UTC') + toIntervalHour(number) AS ts, number % 2 * toDayOfMonth(toDate(ts)) % 2 AS n FROM numbers(105 * 24)")

        def check_result(result):
            self.assertEqual(result[0], {'d': '2021-01-01', 'n': 12})
            self.assertEqual(result[1], {'d': '2021-01-02', 'n': 0})
            self.assertEqual(result[31 + 28 - 1], {'d': '2021-02-28', 'n': 0})
            self.assertEqual(result[31 + 28 + 1 - 1], {'d': '2021-03-01', 'n': 12})
            self.assertEqual(result[-2], {'d': '2021-04-14', 'n': 0})
            self.assertEqual(result[-1], {'d': '2021-04-15', 'n': 12})

        # Once the materialization is done only dates with data are kept
        def check_final_result(result):
            self.assertEqual(len(result), 54)
            self.assertEqual(result[0], {'d': '2021-01-01', 'n': 12})
            self.assertEqual(result[1], {'d': '2021-01-03', 'n': 12})
            self.assertEqual(result[-2], {'d': '2021-04-13', 'n': 12})
            self.assertEqual(result[-1], {'d': '2021-04-15', 'n': 12})

        source_result = exec_sql(self.user.database, f"""
            SELECT
                toDate(ts) d,
                sum(n) as n
            FROM {self.user.database}.{table_source_id}
            GROUP BY d
            ORDER BY d ASC
            FORMAT JSON""", extra_params)['data']

        check_result(source_result)

        view_sql = f"""
            SELECT
                toDate(ts) d,
                sum(n) as n
            FROM {self.user.database}.{table_source_id}
            GROUP BY d
        """
        ch_create_materialized_view_sync(
            self.user.database_server,
            self.user.database,
            view_name=view_node_id,
            sql=view_sql,
            target_table=table_target_id,
            cluster=self.user.cluster
        )
        j = PopulateJob(user=self.user, view_node=view_node_id, view_sql=view_sql, target_table=table_target_id)
        self.job_executor.put_job(j)
        j.save()
        j.run()

        job = await get_finalised_job_async(j.id)

        self.assertEqual(job.status, JobStatus.DONE, job.result)

        # Use FINAL here to ensure we get final results
        target_result = exec_sql(self.user.database, f"""
            SELECT * FROM {table_target_id} FINAL
            ORDER BY d ASC
            FORMAT JSON
        """, extra_params)['data']
        check_final_result(target_result)

    async def test_by_partition_populate_no_partition(self):
        await self._test_by_partition_populate()

    async def test_by_partition_populate_yyyymm_partition(self):
        await self._test_by_partition_populate(partition_key_source='PARTITION BY toYYYYMM(ts)', partition_key_target='PARTITION BY toYYYYMM(d)')


class TestPopulateCancellationFunction(BaseTestJobs):

    def test_returns_false_if_job_not_in_cancelling(self):
        j = PopulateJob(user=self.user, view_node=None, view_sql=None, target_table=None)
        j.save()

        has_been_cancelled_function = j.has_been_externally_cancelled_function_generator()
        self.assertFalse(has_been_cancelled_function())

    def test_returns_true_if_job_in_cancelling(self):
        j = PopulateJob(user=self.user, view_node=None, view_sql=None, target_table=None)
        j.status = JobStatus.CANCELLING
        j.save()

        has_been_cancelled_function = j.has_been_externally_cancelled_function_generator()
        self.assertTrue(has_been_cancelled_function())

    def test_with_the_function_detect_that_a_job_gets_cancelled(self):
        j = PopulateJob(user=self.user, view_node=None, view_sql=None, target_table=None)
        j.save()

        has_been_cancelled_function = j.has_been_externally_cancelled_function_generator()
        self.assertFalse(has_been_cancelled_function())

        j.status = JobStatus.CANCELLING
        j.save()
        self.assertTrue(has_been_cancelled_function())


class ChecksUntilCancellation:
    def __init__(self, checks_remaining):
        self.calls_done = 0
        self.checks_remaining = checks_remaining


class PopulateJobWithCustomCancellable(PopulateJob):

    def __init__(self, user, view_node, view_sql, target_table, checks_until_cancellation):
        self.checks_until_cancellation = checks_until_cancellation
        super().__init__(user, view_node=view_node, view_sql=view_sql, target_table=target_table)

    def has_been_externally_cancelled_function_generator(self) -> Callable[[], bool]:

        def has_been_cancelled() -> bool:
            self.checks_until_cancellation.calls_done += 1

            if self.checks_until_cancellation.checks_remaining == 0:
                self.try_to_cancel()
                self.checks_until_cancellation.checks_remaining -= 1
                return True
            elif self.checks_until_cancellation.checks_remaining < 0:
                return True
            else:
                self.checks_until_cancellation.checks_remaining -= 1
                return False

        return has_been_cancelled


class TestJobsPopulateCancellations(BaseTestJobs):

    async def _create_custom_populate_and_return_it_finished(self, checks_until_cancellation):
        cluster_clause = f"ON CLUSTER {self.user.cluster}" if self.user.cluster else ""
        id = uuid.uuid4().hex
        table_source = f'cancel_test_by_partition_populate_source{id}'
        table_target = f'cancel_test_by_partition_populate_target{id}'
        view_node_id = f'cancel_test_by_partition_populate_mv{id}'

        self.user.drop_datasource(table_source)
        self.user.drop_datasource(table_target)
        table_source_ds = self.user.add_datasource(table_source)
        table_target_ds = self.user.add_datasource(table_target)

        table_source_id = table_source_ds.id
        table_target_id = table_target_ds.id

        tables = [table_source_id, table_target_id, view_node_id]

        tables = [table_source_id, table_target_id, view_node_id]
        for t in tables:
            exec_sql(self.user.database, f"DROP TABLE IF EXISTS {t} {cluster_clause}")

        engine = "MergeTree() PARTITION BY toYYYYMM(ts, 'UTC') ORDER BY tuple()"
        if self.user.cluster:
            engine = engine_local_to_replicated(engine, self.user['database'], table_source_id)
        exec_sql(self.user.database, f"CREATE TABLE {table_source_id} {cluster_clause} (ts DateTime('UTC'), n Int32) Engine = {engine}")

        engine = "SummingMergeTree() PARTITION BY toYYYYMM(d, 'UTC') ORDER BY d"
        if self.user.cluster:
            engine = engine_local_to_replicated(engine, self.user['database'], table_target_id)
        exec_sql(self.user.database, f"CREATE TABLE {table_target_id} {cluster_clause} (d Date, n Int32) Engine = {engine}")

        # Only odd hours of odd days of the month:
        # - 1st of January has 12 hours with data
        # - 4th of February has 0 hours with data
        exec_sql(self.user.database,
                 f"INSERT INTO {table_source_id} SELECT toDateTime('2021-01-01 00:00:00', 'UTC') + toIntervalHour(number) AS ts, number % 2 * toDayOfMonth(toDate(ts)) % 2 AS n FROM numbers(105 * 24)")

        view_sql = f"""
                    SELECT
                        toDate(ts, 'UTC') d,
                        sum(n) as n
                    FROM
                        {self.user.database}.{table_source_id},
                        (SELECT max(number) as number FROM (SELECT number from system.numbers LIMIT 100000000)) as t
                    GROUP BY d
                """
        ch_create_materialized_view_sync(
            self.user.database_server,
            self.user.database,
            view_name=view_node_id,
            sql=view_sql,
            target_table=table_target_id,
            cluster=self.user.cluster
        )
        j = PopulateJobWithCustomCancellable(
            user=self.user, view_node=view_node_id, view_sql=view_sql,
            target_table=table_target_id, checks_until_cancellation=checks_until_cancellation)
        self.job_executor.put_job(j)
        j.save()
        j.run()

        return await wait_until_job_is_in_expected_status_async(j.id, [JobStatus.CANCELLED], max_retries=400)

    async def test_populate_cancellation_may_be_cancelled_inmediately_without_executing_the_queries(self):
        checks_until_cancellation = ChecksUntilCancellation(0)

        job = await self._create_custom_populate_and_return_it_finished(checks_until_cancellation)

        self.assertEqual(job.status, JobStatus.CANCELLED, job.result)
        self.assertEqual(checks_until_cancellation.calls_done, 5)
        queries_cancelled = len(list(filter(lambda query: query.status == 'cancelled', job.queries)))
        self.assertEqual(queries_cancelled, 4)

    async def test_populate_cancellation_may_be_cancelled_and_leave_some_queries_cancelled_and_other_done(self):
        checks_until_cancellation = ChecksUntilCancellation(10)

        job = await self._create_custom_populate_and_return_it_finished(checks_until_cancellation)

        self.assertEqual(job.status, JobStatus.CANCELLED, job.result)
        # It's difficult to get exacts numbers without mocking the checks as it depends on the communication between
        # the backend and CH
        self.assertGreaterEqual(checks_until_cancellation.calls_done, 10)
        self.assertLess(checks_until_cancellation.calls_done, 20)
        queries_cancelled = len(list(filter(lambda query: query.status == 'cancelled', job.queries)))
        self.assertGreaterEqual(queries_cancelled, 1)
        self.assertLess(queries_cancelled, 4)


class TestJobCancellation(BaseTestJobs):
    def setUp(self):
        super().setUp()
        self.job_consumer = self.job_executor.start_consumer()

    def tearDown(self):
        self.job_consumer.terminate()
        self.job_consumer.join()
        super().tearDown()

    def test_job_correctly_set_as_cancelled_from_waiting_status(self):
        j = FakeTestJob(self.user, cancellable_status={JobStatus.WAITING}, block_job_before_executing_it=True)
        j.save()
        self.assertEqual(len(self.job_executor.get_pending_jobs()), 0)
        self.job_executor.put_job(j)

        j.job_already_blocked_before_execution.wait()

        j = Job.get_by_id(j.id)
        self.assertEqual(JobStatus.WAITING, j.status)

        pending_job_ids = [job.id for job in self.job_executor.get_pending_jobs()]
        self.assertIn(j.id, pending_job_ids)

        j.try_to_cancel()
        j = Job.get_by_id(j.id)
        self.assertEqual(JobStatus.CANCELLED, j.status)

        self.assertEqual(len(self.job_executor.get_pending_jobs()), 0)
        j.before_execution_event.set()  # Unblock the blocked job

    def test_job_correctly_set_as_cancelled_from_cancelling_status(self):
        j = FakeTestJob(self.user, cancellable_status={JobStatus.WAITING, JobStatus.WORKING})
        j.save()
        self.assertEqual(len(self.job_executor.get_pending_jobs()), 0)
        self.job_executor.put_job(j)

        j = j.wait_until_working()
        j.try_to_cancel()
        j = Job.get_by_id(j.id)
        self.assertEqual(JobStatus.CANCELLING, j.status)
        pending_job_ids = [job.id for job in self.job_executor.get_pending_jobs()]
        self.assertIn(j.id, pending_job_ids)

        j.notify_cancelled_event()
        j.wait_until_finalised()

        j = Job.get_by_id(j.id)
        self.assertEqual(JobStatus.CANCELLED, j.status)
        self.assertEqual(len(self.job_executor.get_pending_jobs()), 0)


class TestJobOwnership(BaseTestJobs):

    def test_job_is_owned_by_user(self):
        j = PopulateJob(user=self.user, view_node="", view_sql="", target_table="")
        j.save()

        is_owned = Job.is_owned_by(j.id, self.user.id)
        self.assertTrue(is_owned)

    def test_job_is_not_owned_by_other_user(self):
        j = PopulateJob(user=self.user, view_node="", view_sql="", target_table="")
        j.save()

        is_owned = Job.is_owned_by(j.id, "other_user_id")
        self.assertFalse(is_owned)


class TestPendingJobs(BaseTestJobs):

    def test_job_in_pending(self):
        j = FakeTestJob(self.user)
        j.save()
        consumer = self.job_executor.start_consumer()
        try:
            self.assertEqual(j.status, JobStatus.WORKING)

            self.assertEqual(j.status, 'waiting')
            self.job_executor.put_job(j)

            j = j.wait_until_working()

            pending_job_ids = [job.id for job in self.job_executor.get_pending_jobs()]
            self.assertIn(j.id, pending_job_ids)

            j.finish({'task': 'ok'})

            job = j.wait_until_finalised()
            self.assertEqual(job.status, 'done')
            self.assertEqual(job.result['task'], 'ok')

            pending_job_ids = [job.id for job in self.job_executor.get_pending_jobs()]
            self.assertNotIn(j.id, pending_job_ids)
        except Exception:
            consumer.terminate()
            consumer.join()

    def test_pending_and_wip_and_queued_jobs(self):
        j = FakeTestJob(self.user)
        j.set_do_not_submit()
        j.save()

        self.assertEqual(j.status, 'waiting')
        self.job_executor.put_job(j)

        wip_jobs, queued_jobs = self.job_executor.get_wip_and_queued_jobs()
        pending_jobs = self.job_executor.get_pending_jobs()

        self.assertEqual(pending_jobs, wip_jobs + queued_jobs)

    def _common_pending_test(self, status, in_pending=True):
        j = FakeTestJob(self.user)
        j.set_do_not_submit()
        j.save()

        self.assertEqual(j.status, 'waiting')
        self.job_executor.put_job(j)

        wip_jobs = []
        queued_jobs = []
        pending_jobs = []

        def _update_pending_jobs():
            nonlocal wip_jobs, queued_jobs, pending_jobs
            wip_jobs, queued_jobs = self.job_executor.get_wip_and_queued_jobs()
            pending_jobs = self.job_executor.get_pending_jobs()
            self.assertEqual(pending_jobs, wip_jobs + queued_jobs)
            wip_jobs = [job.id for job in wip_jobs]
            queued_jobs = [job.id for job in queued_jobs]
            pending_jobs = [job.id for job in pending_jobs]

        _update_pending_jobs()
        self.assertNotIn(j.id, wip_jobs)
        self.assertIn(j.id, queued_jobs)
        self.assertIn(j.id, pending_jobs)

        j.status = status
        j.save()
        self.assertEqual(j.status, status)

        _update_pending_jobs()
        # Jobs in working status that have not been picked up to be run yet (and
        # thus, added to the WIP queue) shouldn't be present as a pending job.
        if in_pending and j.status != JobStatus.WORKING:
            self.assertNotIn(j.id, wip_jobs)
            self.assertIn(j.id, queued_jobs)
            self.assertIn(j.id, pending_jobs)
        else:
            self.assertNotIn(j.id, wip_jobs)
            self.assertNotIn(j.id, queued_jobs)
            self.assertNotIn(j.id, pending_jobs)

        # Take out of the Redis queue the job
        executors = self.job_executor.get_all_executors()
        self.assertEqual(len(executors), 1)
        job_id = executors[0].get_job()
        self.assertEqual(job_id, j.id)

        _update_pending_jobs()
        if in_pending:
            self.assertIn(j.id, wip_jobs)
            self.assertNotIn(j.id, queued_jobs)
            self.assertIn(j.id, pending_jobs)
        else:
            self.assertNotIn(j.id, wip_jobs)
            self.assertNotIn(j.id, queued_jobs)
            self.assertNotIn(j.id, pending_jobs)

    def test_waiting_job_in_pending(self):
        self._common_pending_test(JobStatus.WAITING, in_pending=True)

    def test_cancelling_job_in_pending(self):
        self._common_pending_test(JobStatus.CANCELLING, in_pending=True)

    def test_working_job_in_pending(self):
        self._common_pending_test(JobStatus.WORKING, in_pending=True)

    def test_finished_job_not_in_pending(self):
        self._common_pending_test(JobStatus.DONE, in_pending=False)

    def test_cancelled_job_not_in_pending(self):
        self._common_pending_test(JobStatus.CANCELLED, in_pending=False)

    def test_errored_job_not_in_pending(self):
        self._common_pending_test(JobStatus.ERROR, in_pending=False)

    def test_removed_job_not_in_pending(self):
        j1 = FakeTestJob(self.user)
        j1.set_do_not_submit()
        j2 = FakeTestJob(self.user)
        j2.set_do_not_submit()
        j3 = FakeTestJob(self.user)
        j3.set_do_not_submit()

        self.job_executor.put_job(j1)
        self.job_executor.put_job(j2)
        self.job_executor.put_job(j3)
        pending_job_ids = [job.id for job in self.job_executor.get_pending_jobs()]
        self.assertIn(j1.id, pending_job_ids)
        self.assertIn(j2.id, pending_job_ids)
        self.assertIn(j3.id, pending_job_ids)

        # Remove the job j2 from Redis to check that it's no longer collected as a pending job
        Job._delete(j2.id)

        # Take out of the queued all jobs so that they're added into the WIP set
        for executor in self.job_executor.get_all_executors():
            # Note that since they share the same user, they share the same database server and all 3 jobs fall
            # into the same executor, so ask it 3 times to get jobs
            _ = executor.get_job()
            _ = executor.get_job()
            _ = executor.get_job()

        pending_job_ids = [job.id for job in self.job_executor.get_pending_jobs()]
        self.assertIn(j1.id, pending_job_ids)
        self.assertNotIn(j2.id, pending_job_ids)
        self.assertIn(j3.id, pending_job_ids)




import os
import unittest
from tinybird.datafile import parse, ValidationException, ParseException


class TestDatafile(unittest.TestCase):

    def setUp(self):
        super().setUp()

    def test_parse_version(self):
        s = """
        VERSION 10
        """
        doc = parse(s)
        self.assertEqual(doc.version, 10)

    def test_parse_wrong_version(self):
        s = """
        VERSION -1
        """
        with self.assertRaises(ValidationException):
            parse(s)

    def test_general_parse(self):
        s = """
FROM SCRATCH
SOURCE 'https://example.com'
# this is a comment
MAINTAINER 'rambo' # this is me
NODE test_01
DESCRIPTION this is a node that does whatever
SQL >
  SELECT * from test_00

NODE test_02
SQL >
   SELECT * from test_01
   WHERE a > 1
   GROUP by a
        """
        d = parse(s)
        self.assertEqual(d.maintainer, 'rambo')
        self.assertEqual(len(d.nodes), 2, d.nodes)
        self.assertEqual(d.nodes[0], {
            'name': 'test_01',
            'description': 'this is a node that does whatever', 'sql': 'SELECT * from test_00'
        })
        self.assertEqual(d.nodes[1], {'name': 'test_02', 'sql': 'SELECT * from test_01\nWHERE a > 1\nGROUP by a'})

    def test_pipe_description_parse(self):
        content = """
DESCRIPTION >
    This is the main pipe description
NODE test_01
DESCRIPTION this is the first node description
SQL >
    SELECT * from test_00
        """
        pipe = parse(content)
        self.assertEqual(pipe.description, 'This is the main pipe description')
        self.assertEqual(len(pipe.nodes), 1, pipe.nodes)

    def test_datasource_description_parse(self):
        content = """
DESCRIPTION >
    This is the main datasource description
        """
        datasource = parse(content, 'default')
        self.assertEqual(datasource.nodes[0]['description'], 'This is the main datasource description')

    def test_datasource_engine(self):
        s = """VERSION 0
SCHEMA >
    a Date,
    b Float32,
    c String

ENGINE "MergeTree"
ENGINE_PARTITION_KEY "toYYYYMM(a)"
ENGINE_SORTING_KEY "(c, a)"
        """
        ds = parse(s, 'default')
        self.assertIn('engine', ds.nodes[0])
        self.assertEqual(ds.nodes[0]['engine']['type'], 'MergeTree')
        self.assertEqual(ds.nodes[0]['engine']['args'], [('partition_key', 'toYYYYMM(a)'), ('sorting_key', '(c, a)')])

    def test_parse_engine_no_node(self):
        s = """
        ENGINE "MergeTree"
        """
        with self.assertRaisesRegex(Exception, "ENGINE must be called after a NODE command"):
            parse(s)

        s = """
        ENGINE_PARTITION_KEY "toYYYYMM(a)"
        """
        with self.assertRaisesRegex(Exception, "ENGINE_PARTITION_KEY must be called after a NODE command"):
            parse(s)

    def test_node_engine(self):
        s = """VERSION 0
NODE mat_node
SQL >
    SELECT toDate(ts) as date, country, sum(sales) sum_sales FROM ds

DATASOURCE "daily_sales_by_country"
ENGINE "SummingMergeTree"
ENGINE_PARTITION_KEY "toYYYYMM(date)"
ENGINE_SORTING_KEY "(country, date)"
ENGINE_SETTINGS "index_granularity=32"
ENGINE_TTL "toDate(date) + INTERVAL 3 DAY"
        """
        ds = parse(s)
        self.assertIn('engine', ds.nodes[0])
        self.assertEqual(ds.nodes[0]['engine']['type'], 'SummingMergeTree')
        self.assertEqual(ds.nodes[0]['engine']['args'], [('partition_key', 'toYYYYMM(date)'), ('sorting_key', '(country, date)'), ('settings', 'index_granularity=32'), ('ttl', 'toDate(date) + INTERVAL 3 DAY')])
        self.assertEqual(ds.nodes[0]['datasource'], 'daily_sales_by_country')

    def test_no_indentation_parse(self):
        s = """VERSION 0
NODE mat_node
SQL >
SELECT toDate(ts) as date, country, sum(sales) sum_sales FROM ds

DATASOURCE "daily_sales_by_country"
ENGINE "SummingMergeTree"
ENGINE_PARTITION_KEY "toYYYYMM(date)"
ENGINE_SORTING_KEY "(country, date)"
ENGINE_SETTINGS "index_granularity=32"
ENGINE_TTL "toDate(date) + INTERVAL 3 DAY"
        """
        ds = parse(s)
        self.assertIn('engine', ds.nodes[0])
        self.assertEqual(ds.nodes[0]['sql'], 'SELECT toDate(ts) as date, country, sum(sales) sum_sales FROM ds')
        self.assertEqual(ds.nodes[0]['engine']['type'], 'SummingMergeTree')
        self.assertEqual(ds.nodes[0]['engine']['args'], [('partition_key', 'toYYYYMM(date)'), ('sorting_key', '(country, date)'), ('settings', 'index_granularity=32'), ('ttl', 'toDate(date) + INTERVAL 3 DAY')])
        self.assertEqual(ds.nodes[0]['datasource'], 'daily_sales_by_country')

    def test_no_indentation_multiline_parse(self):
        s = """VERSION 0
NODE mat_node
SQL >
SELECT toDate(ts) as date,
country,
sum(sales) sum_sales
FROM ds

DATASOURCE "daily_sales_by_country"
ENGINE "SummingMergeTree"
ENGINE_PARTITION_KEY "toYYYYMM(date)"
ENGINE_SORTING_KEY "(country, date)"
ENGINE_SETTINGS "index_granularity=32"
ENGINE_TTL "toDate(date) + INTERVAL 3 DAY"
        """
        ds = parse(s)
        self.assertIn('engine', ds.nodes[0])
        self.assertEqual(ds.nodes[0]['sql'], 'SELECT toDate(ts) as date,\ncountry,\nsum(sales) sum_sales\nFROM ds')
        self.assertEqual(ds.nodes[0]['engine']['type'], 'SummingMergeTree')
        self.assertEqual(ds.nodes[0]['engine']['args'], [('partition_key', 'toYYYYMM(date)'), ('sorting_key', '(country, date)'), ('settings', 'index_granularity=32'), ('ttl', 'toDate(date) + INTERVAL 3 DAY')])
        self.assertEqual(ds.nodes[0]['datasource'], 'daily_sales_by_country')

    def test_no_indentation_multiline_group_parse(self):
        s = """VERSION 0
NODE mat_node
SQL >
SELECT toDate(ts) as date,
country,
sum(sales) sum_sales
FROM ds
GROUP BY date
country

DATASOURCE "daily_sales_by_country"
ENGINE "SummingMergeTree"
ENGINE_PARTITION_KEY "toYYYYMM(date)"
ENGINE_SORTING_KEY "(country, date)"
ENGINE_SETTINGS "index_granularity=32"
ENGINE_TTL "toDate(date) + INTERVAL 3 DAY"
        """
        ds = parse(s)
        self.assertIn('engine', ds.nodes[0])
        self.assertEqual(ds.nodes[0]['sql'], 'SELECT toDate(ts) as date,\ncountry,\nsum(sales) sum_sales\nFROM ds\nGROUP BY date\ncountry')
        self.assertEqual(ds.nodes[0]['engine']['type'], 'SummingMergeTree')
        self.assertEqual(ds.nodes[0]['engine']['args'], [('partition_key', 'toYYYYMM(date)'), ('sorting_key', '(country, date)'), ('settings', 'index_granularity=32'), ('ttl', 'toDate(date) + INTERVAL 3 DAY')])
        self.assertEqual(ds.nodes[0]['datasource'], 'daily_sales_by_country')

    def test_no_indentation_multiline_with_validation_error_parse(self):
        s = """VERSION 0
NODE mat_node
SQL >
SELECT toDate(ts) as date, country, sum(sales) sum_sales FROM test
as

DATASOURCE
where date = '2020-01-01'

DATASOURCE "test"
        """
        with self.assertRaises(ValidationException) as ctx:
            parse(s)
        self.assertEqual("Validation error, found DATASOURCE\n in line 6: string index out of range", str(ctx.exception))

    def test_no_indentation_multiline_with_parse_error(self):
        s = """VERSION 0
NODE mat_node
SQL >
SELECT toDate(ts) as date, country, sum(sales) sum_sales FROM test
as

DATASOURCE "test"
where date = '2020-01-01'
        """
        with self.assertRaises(ParseException) as ctx:
            parse(s)
        self.assertIn('WHERE is not a valid option', str(ctx.exception))

    def test_read_env_vars(self):
        s = """VERSION 0
NODE "mat_${test}"

SQL >
SELECT toDate(ts) as date,
country,
sum(sales) sum_sales
FROM ds
GROUP BY date
country

DATASOURCE "${test}_ds"
ENGINE "SummingMergeTree${doesnotexist}"
        """
        os.environ['test'] = 'rambo'
        ds = parse(s)
        self.assertIn('datasource', ds.nodes[0])
        self.assertEqual(ds.nodes[0]['datasource'], 'rambo_ds')
        self.assertEqual(ds.nodes[0]['name'], 'mat_rambo')
        self.assertEqual(ds.nodes[0]['engine']['type'], 'SummingMergeTree${doesnotexist}')

    def test_parse_datasource_import(self):
        content = """
DESCRIPTION >
    Test

SCHEMA >
    `Year` Nullable(Int64),
    `Industry_aggregation_NZSIOC` Nullable(String),
    `Industry_code_NZSIOC` Nullable(String),
    `Industry_name_NZSIOC` Nullable(String),
    `Units` Nullable(String),
    `Variable_code` Nullable(String),
    `Variable_name` Nullable(String),
    `Variable_category` Nullable(String),
    `Value` Nullable(String),
    `Industry_code_ANZSIC06` Nullable(String)

ENGINE "MergeTree"
ENGINE_PARTITION_KEY "toYear(insertion_date)"
ENGINE_SORTING_KEY "insertion_date"

IMPORT_SERVICE bigquery
IMPORT_SCHEDULE "*/25 * * * *"
IMPORT_STRATEGY replace
IMPORT_QUERY >
    SELECT * FROM avid-atlas-375013.test_data.test_table
        """
        ds = parse(content, 'test')

        node = ds.nodes[0]
        self.assertEqual(node['import_service'], 'bigquery')
        self.assertEqual(node['import_schedule'], '*/25 * * * *')
        self.assertEqual(node['import_strategy'], 'replace')
        self.assertEqual(node['import_query'].strip(), 'SELECT * FROM avid-atlas-375013.test_data.test_table')

    def test_parse_datasource_import_strip_query(self):
        content = """
DESCRIPTION >
    Test

SCHEMA >
    `Year` Nullable(Int64)

ENGINE "MergeTree"
ENGINE_PARTITION_KEY "toYear(insertion_date)"
ENGINE_SORTING_KEY "insertion_date"

IMPORT_SERVICE bigquery
IMPORT_SCHEDULE "*/25 * * * *"
IMPORT_STRATEGY replace
IMPORT_QUERY 'SELECT * FROM avid-atlas-375013.test_data.test_table'
        """
        ds = parse(content, 'test')

        node = ds.nodes[0]
        self.assertEqual(node['import_query'].strip(), 'SELECT * FROM avid-atlas-375013.test_data.test_table')

    def test_parse_datasource_import_strip_multiline_query(self):
        content = """
DESCRIPTION >
    Test

SCHEMA >
    `Year` Nullable(Int64)

ENGINE "MergeTree"
ENGINE_PARTITION_KEY "toYear(insertion_date)"
ENGINE_SORTING_KEY "insertion_date"

IMPORT_SERVICE bigquery
IMPORT_SCHEDULE "*/25 * * * *"
IMPORT_STRATEGY replace
IMPORT_QUERY >
    'SELECT * FROM avid-atlas-375013.test_data.test_table'
        """
        ds = parse(content, 'test')

        node = ds.nodes[0]
        self.assertEqual(node['import_query'].strip(), 'SELECT * FROM avid-atlas-375013.test_data.test_table')



import unittest

from tinybird.sampler import guess


class TestSampler(unittest.TestCase):

    def test_guess_simple(self):
        self.maxDiff = None
        guessed = []
        guess('my_user', 'my_ds', guessed, {'x': 3, 'y': 'wadus'}, 123)
        self.assertEqual([
            ('my_user', 'my_ds', 123, '$', 'object', 0, ''),
            ('my_user', 'my_ds', 123, '$.x', 'number', 3, ''),
            ('my_user', 'my_ds', 123, '$.y', 'string', 0, 'wadus')
        ], guessed)

    def test_guess_nested(self):
        self.maxDiff = None
        guessed = []
        guess('my_user', 'my_ds', guessed, {'x': {'y': 4}}, 123)
        self.assertEqual([
            ('my_user', 'my_ds', 123, '$', 'object', 0, ''),
            ('my_user', 'my_ds', 123, '$.x', 'object', 0, ''),
            ('my_user', 'my_ds', 123, '$.x.y', 'number', 4, '')
        ], guessed)



from typing import Any, Dict, List
import unittest.mock
import uuid

import tornado.testing

from tests.conftest import get_redis_client
from tinybird.auth0_client import Auth0Client

from tinybird.useraccounts_service import UserAccountsService

from tinybird.user import UserAccount


def patch_get_users_by_email(data: Dict[str, Any]) -> Dict[str, Any]:
    return unittest.mock.patch.object(Auth0Client, 'get_users_by_email', return_value=data)


class TestUserAccountsService(tornado.testing.AsyncTestCase):

    def setUp(self):
        super(TestUserAccountsService, self).setUp()
        self.users_to_delete: List[UserAccount] = []
        self.redis_client = get_redis_client()

    def tearDown(self):
        for u in self.users_to_delete:
            UserAccount._delete(u.id)
        super(TestUserAccountsService, self).tearDown()

    def _create_user(self, email: str = None) -> UserAccount:
        email = email if email else f'test_{uuid.uuid4().hex}@example.com'
        user = UserAccount.register(email, 'pass')
        self.users_to_delete.append(user)
        return user

    @tornado.testing.gen_test
    async def test_get_auth_provider_info(self):
        new_user = self._create_user()

        with patch_get_users_by_email([{'logins_count': 4}, {'logins_count': 1}]):
            info = await UserAccountsService.get_auth_provider_info(new_user)
            self.assertEqual(info['logins_count'], 5)

        with patch_get_users_by_email([{'logins_count': 0}, {'logins_count': 0}]):
            info = await UserAccountsService.get_auth_provider_info(new_user)
            self.assertEqual(info['logins_count'], 0)

        with patch_get_users_by_email([{'logins_count': 42}]):
            info = await UserAccountsService.get_auth_provider_info(new_user)
            self.assertEqual(info['logins_count'], 42)

        with patch_get_users_by_email([]):
            info = await UserAccountsService.get_auth_provider_info(new_user)
            self.assertEqual(info['logins_count'], 0)



import asyncio
from typing import Any, Dict, Iterable, Optional
import tornado
import uuid

from tinybird.organization.organization import Organization, OrganizationCommitmentsPlans, Organizations
from tinybird.user import UserAccount, User, UserAccountAlreadyBelongsToOrganization, WorkspaceAlreadyBelongsToOrganization

from .views.base_test import BaseTest


class TestOrganizations(BaseTest):
    def setUp(self) -> None:
        self.organizations_to_delete = []
        super().setUp()

    def tearDown(self) -> None:
        for org in self.organizations_to_delete:
            asyncio.run(Organizations.remove_organization(org))
        super().tearDown()

    def _create_user(self, prefix: str) -> UserAccount:
        email: str = f'{prefix}_{uuid.uuid4().hex}@example.com'
        result: UserAccount = UserAccount.register(email, 'pass')
        self.users_to_delete.append(result)
        return result

    def _create_workspace(self, prefix: str, user: UserAccount) -> User:
        workspace_name: str = f'workspace_{prefix}_{uuid.uuid4().hex}'
        result: User = User.register(workspace_name, user.id)
        self.workspaces_to_delete.append(result)
        return result

    async def _create_organization(self, name: str, domain: Optional[str] = None,
                                   plan_details: Optional[Dict[str, Any]] = None,
                                   workspace_ids: Optional[Iterable[str]] = None,
                                   user_ids: Optional[Iterable[str]] = None) -> Organization:
        org = Organization.create(name)
        self.organizations_to_delete.append(org)

        if domain:
            org = await Organizations.update_name_and_domain(org, name, domain)

        if plan_details:
            org = await Organizations.update_commitment_information(
                org,
                start_date=plan_details['start_date'],
                end_date=plan_details['end_date'],
                commited_processed=plan_details['commitment']['processed'],
                commited_storage=plan_details['commitment']['storage'],
                commitment_billing=plan_details['billing']
            )

        if workspace_ids:
            for id in workspace_ids:
                org = await Organizations.add_workspace(org, User.get_by_id(id))

        if user_ids:
            for id in user_ids:
                org = await Organizations.add_user(org, UserAccount.get_by_id(id))

        return org

    @tornado.testing.gen_test
    async def test_create_organization(self) -> None:
        """Tests a new Organization defaults.
        """
        user = self._create_user('test_create_organization')
        workspace = self._create_workspace('test_create_organization', user)
        org = await self._create_organization('test_create_organization',
                                              user_ids=[user.id],
                                              workspace_ids=[workspace.id])
        # Linked users and workspaces
        self.assertEqual(UserAccount.get_by_id(user.id).organization_id, org.id)
        self.assertEqual(User.get_by_id(workspace.id).organization_id, org.id)

        # Metadata
        self.assertEqual(org.name, 'test_create_organization')
        self.assertEqual(org.domain, None)
        self.assertEqual(org.plan_details['name'], '')
        self.assertNotEqual(org.plan_details['start_date'], '')
        self.assertEqual(org.plan_details['end_date'], '')
        self.assertEqual(org.plan_details['commitment']['storage'], 0)
        self.assertEqual(org.plan_details['commitment']['processed'], 0)
        self.assertEqual(org.plan_details['billing'], OrganizationCommitmentsPlans.TOTAL_USAGE)

    @tornado.testing.gen_test
    async def test_remove_organization(self) -> None:
        """Test the removal of an organization successfully unlinks
           all linked users and workspaces.
        """
        user = self._create_user('test_create_organization')
        workspace = self._create_workspace('test_create_organization', user)
        org = await self._create_organization('test_create_organization',
                                              user_ids=[user.id],
                                              workspace_ids=[workspace.id])
        await Organizations.remove_organization(org)

        self.assertEqual(UserAccount.get_by_id(user.id).organization_id, None)
        self.assertEqual(User.get_by_id(workspace.id).organization_id, None)

    @tornado.testing.gen_test
    async def test_update_meta(self) -> None:
        """Test the metadata update methods.
        """
        org = await self._create_organization('test_update_meta')

        org = await Organizations.update_name_and_domain(org, 'new_name', 'example.com')
        self.assertEqual(org.name, 'new_name')
        self.assertEqual(org.domain, 'example.com')

        org = await Organizations.update_commitment_information(org, '2100-01-01', '2100-12-31',
                                                                999, 666,
                                                                OrganizationCommitmentsPlans.NO_USAGE_COMMITMENT)
        self.assertEqual(org.plan_details['start_date'], '2100-01-01')
        self.assertEqual(org.plan_details['end_date'], '2100-12-31')
        self.assertEqual(org.plan_details['commitment']['storage'], 666)
        self.assertEqual(org.plan_details['commitment']['processed'], 999)
        self.assertEqual(org.plan_details['billing'], OrganizationCommitmentsPlans.NO_USAGE_COMMITMENT)

    @tornado.testing.gen_test
    async def test_add_workspace(self) -> None:
        """Tests we can't add the same workspace to two organizations.
        """
        user = self._create_user('test_add_workspace')
        workspace = self._create_workspace('test_add_workspace', user)
        org = await self._create_organization('test_add_workspace', workspace_ids=[workspace.id])
        self.assertEqual(len(org.workspace_ids), 1)

        try:
            _ = await self._create_organization('test_add_workspace_2', workspace_ids=[workspace.id])
            self.assertTrue(False, "We can't add a workspace to two organizations")
        except WorkspaceAlreadyBelongsToOrganization as ex:
            self.assertEqual(ex.workspace_id, workspace.id)
            self.assertEqual(ex.organization_id, org.id)

    @tornado.testing.gen_test
    async def test_remove_workspace(self) -> None:
        """Tests the removal of a workspace.
        """
        user = self._create_user('test_add_workspace')
        workspace = self._create_workspace('test_add_workspace', user)
        org = await self._create_organization('test_add_workspace', workspace_ids=[workspace.id])
        self.assertEqual(len(org.workspace_ids), 1)
        org = await Organizations.remove_workspace(org, workspace)
        self.assertEqual(len(org.workspace_ids), 0)
        self.assertEqual(User.get_by_id(workspace.id).organization_id, None)

    @tornado.testing.gen_test
    async def test_add_user(self) -> None:
        """Tests we can't add the same user to two organizations.
        """
        user = self._create_user('test_add_user')
        org = await self._create_organization('test_add_user', user_ids=[user.id])
        self.assertEqual(len(org.user_account_ids), 1)

        try:
            _ = await self._create_organization('test_add_user_2', user_ids=[user.id])
            self.assertTrue(False, "We can't add an user to two organizations")
        except UserAccountAlreadyBelongsToOrganization as ex:
            self.assertEqual(ex.user_id, user.id)
            self.assertEqual(ex.organization_id, org.id)

    @tornado.testing.gen_test
    async def test_remove_user(self) -> None:
        """Tests the removal of an user.
        """
        user = self._create_user('test_remove_user')
        org = await self._create_organization('test_remove_user', user_ids=[user.id])
        self.assertEqual(len(org.user_account_ids), 1)
        org = await Organizations.remove_user(org, user)
        self.assertEqual(len(org.user_account_ids), 0)
        self.assertEqual(UserAccount.get_by_id(user.id).organization_id, None)

    @tornado.testing.gen_test
    async def test_refresh_token(self) -> None:
        """Tests the observability token refresh.
        """
        user = self._create_user('test_refresh_token')
        org = await self._create_organization('test_refresh_token', user_ids=[user.id])
        token = org.token_observability.token
        org, refreshed = await Organizations.refresh_token(org, token)
        self.assertNotEqual(token, refreshed)

    @tornado.testing.gen_test
    async def test_refresh_invalid_token(self) -> None:
        """Tests the observability token is NOT refreshed by error.
        """
        user = self._create_user('test_refresh_invalid_token')
        org = await self._create_organization('test_refresh_invalid_token', user_ids=[user.id])
        token = org.token_observability.token
        org, refreshed = await Organizations.refresh_token(org, 'this_token_does_not_exist')
        self.assertEqual(refreshed, None)
        self.assertEqual(token, org.token_observability.token)



import unittest
import uuid
import time
import asyncio

from tinybird.redis_client import get_redis_client, get_redis_args
from tinybird.job import JobExecutor, Job, JobKind, JobStatus
from tinybird.user import User, UserAccount
from tinybird.ch import HTTPClient
from .utils import get_finalised_job_async


PYTHON_THREAD_TOLERANCE_IN_SECONDS = 2


class FakeTestJob(Job):
    def __init__(self, kind, user, job_time=None):
        Job.__init__(self, kind, user)
        self.database_server = user['database_server']
        self.job_time = job_time
        self.save()

    def run(self):
        def function_to_execute(job):
            if self.job_time:
                time.sleep(self.job_time)
            job.status = JobStatus.DONE
            job.save()
            self.job_executor.job_finished(job)

        self.job_executor.submit(function_to_execute, self)


class TestJobExecutor(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        super().setUp()

        redis_kwargs = get_redis_args()

        self.user, self.user_account = self.new_user()

        self.client = HTTPClient(self.user.database_server, database=None)
        self.client.query_sync(f"CREATE DATABASE IF NOT EXISTS {self.user.database} ON CLUSTER tinybird", read_only=False)

        JobExecutor.CHECK_NEW_QUEUES_FREQ_IN_SECONDS = 0.05
        JobExecutor.ALIVE_QUEUES_TTL_IN_SECONDS = 1
        redis_client = get_redis_client()
        self.job_executor_producer = JobExecutor(redis_client=redis_client, redis_kwargs=redis_kwargs, consumer=False, import_workers=0, query_workers=0, export_workers=0, branching_workers=0)
        self.job_executor_consumer = JobExecutor(redis_client=redis_client, redis_kwargs=redis_kwargs, consumer=True, import_workers=1, query_workers=1, export_workers=1, branching_workers=1)
        self.job_consumer = self.job_executor_consumer.start_consumer()

    def tearDown(self):
        self.job_consumer.terminate()
        self.job_consumer.join()
        self.job_executor_producer.join(wait=True)
        self.job_executor_consumer.join(wait=True)
        self.job_executor_consumer._clean()

        self.delete_user(self.user, self.user_account)
        super().tearDown()

    def new_user(self):
        WORKSPACE = f'test_job_executor_{str(int(time.time()))}_{uuid.uuid4().hex}'
        USER = f'{WORKSPACE}@example.com'

        try:
            user_account = UserAccount.get_by_email(USER)
            UserAccount._delete(user_account.id)
        except Exception:
            pass

        user_account = UserAccount.register(USER, 'pass')
        user = User.register(
            name=WORKSPACE,
            admin=user_account.id
        )

        return user, user_account

    def delete_user(self, user, user_account):
        User._delete(user.id)
        UserAccount._delete(user_account.id)
        self.client.query_sync(f"DROP DATABASE IF EXISTS `{user.database}` ON CLUSTER tinybird", read_only=False)

    async def check_threadpool_exists(self, threadpool):
        start = time.time()
        while (time.time() - start) < (JobExecutor.ALIVE_QUEUES_TTL_IN_SECONDS + PYTHON_THREAD_TOLERANCE_IN_SECONDS):
            if self.user.database_server in threadpool:
                return
            await asyncio.sleep(JobExecutor.CHECK_NEW_QUEUES_FREQ_IN_SECONDS)
        self.assertTrue(False)

    async def check_discover_new_queue(self, kind):
        job = FakeTestJob(kind, self.user)
        self.job_executor_producer.put_job(job)

        if kind == JobKind.IMPORT:
            self.assertTrue(self.user.database_server in self.job_executor_producer._import_threadpool_executors)
            self.assertEqual(self.job_executor_producer._import_threadpool_executors[self.user.database_server]._thread_pool_executor, None)
            self.assertEqual(len(self.job_executor_producer._query_threadpool_executors), 0)
            self.assertEqual(len(self.job_executor_consumer._export_threadpool_executors), 0)
            self.assertEqual(len(self.job_executor_consumer._branching_threadpool_executors), 0)
            await self.check_threadpool_exists(self.job_executor_consumer._import_threadpool_executors)
            self.assertEqual(len(self.job_executor_consumer._query_threadpool_executors), 0)
            self.assertEqual(len(self.job_executor_consumer._export_threadpool_executors), 0)
            self.assertEqual(len(self.job_executor_consumer._branching_threadpool_executors), 0)
        elif kind == JobKind.COPY:
            self.assertTrue(self.user.database_server in self.job_executor_producer._export_threadpool_executors)
            self.assertEqual(self.job_executor_producer._export_threadpool_executors[self.user.database_server]._thread_pool_executor, None)
            self.assertEqual(len(self.job_executor_consumer._query_threadpool_executors), 0)
            self.assertEqual(len(self.job_executor_consumer._import_threadpool_executors), 0)
            self.assertEqual(len(self.job_executor_consumer._branching_threadpool_executors), 0)
            await self.check_threadpool_exists(self.job_executor_consumer._export_threadpool_executors)
            self.assertEqual(len(self.job_executor_consumer._query_threadpool_executors), 0)
            self.assertEqual(len(self.job_executor_consumer._import_threadpool_executors), 0)
            self.assertEqual(len(self.job_executor_consumer._branching_threadpool_executors), 0)
        elif kind == JobKind.REGRESSION:
            self.assertTrue(self.user.database_server in self.job_executor_producer._branching_threadpool_executors)
            self.assertEqual(self.job_executor_producer._branching_threadpool_executors[self.user.database_server]._thread_pool_executor, None)
            self.assertEqual(len(self.job_executor_consumer._query_threadpool_executors), 0)
            self.assertEqual(len(self.job_executor_consumer._import_threadpool_executors), 0)
            self.assertEqual(len(self.job_executor_consumer._export_threadpool_executors), 0)
            await self.check_threadpool_exists(self.job_executor_consumer._branching_threadpool_executors)
            self.assertEqual(len(self.job_executor_consumer._query_threadpool_executors), 0)
            self.assertEqual(len(self.job_executor_consumer._import_threadpool_executors), 0)
            self.assertEqual(len(self.job_executor_consumer._export_threadpool_executors), 0)
        else:
            self.assertTrue(self.user.database_server in self.job_executor_producer._query_threadpool_executors)
            self.assertEqual(self.job_executor_producer._query_threadpool_executors[self.user.database_server]._thread_pool_executor, None)
            self.assertEqual(len(self.job_executor_consumer._query_threadpool_executors), 0)
            self.assertEqual(len(self.job_executor_consumer._branching_threadpool_executors), 0)
            await self.check_threadpool_exists(self.job_executor_consumer._query_threadpool_executors)
            self.assertEqual(len(self.job_executor_consumer._import_threadpool_executors), 0)
            self.assertEqual(len(self.job_executor_consumer._export_threadpool_executors), 0)
            self.assertEqual(len(self.job_executor_consumer._branching_threadpool_executors), 0)

    async def check_threadpool_executors_is_empty(self, threadpool):
        start = time.time()
        while (time.time() - start) < (JobExecutor.ALIVE_QUEUES_TTL_IN_SECONDS + JobExecutor.CHECK_NEW_QUEUES_FREQ_IN_SECONDS + 2 * PYTHON_THREAD_TOLERANCE_IN_SECONDS):
            if len(threadpool) == 0:
                return
            await asyncio.sleep(JobExecutor.CHECK_NEW_QUEUES_FREQ_IN_SECONDS)
        self.assertTrue(False)

    async def test_discover_new_import_queue(self):
        await self.check_discover_new_queue(JobKind.IMPORT)

    async def test_discover_new_query_queue(self):
        await self.check_discover_new_queue(JobKind.POPULATE)

    async def test_discover_new_export_queue(self):
        await self.check_discover_new_queue(JobKind.COPY)

    async def test_import_workers_shutdown(self):
        await self.check_discover_new_queue(JobKind.IMPORT)
        await self.check_threadpool_executors_is_empty(self.job_executor_consumer._import_threadpool_executors)

    async def test_query_workers_shutdown(self):
        await self.check_discover_new_queue(JobKind.QUERY)
        await self.check_threadpool_executors_is_empty(self.job_executor_consumer._query_threadpool_executors)

    async def test_export_workers_shutdown(self):
        await self.check_discover_new_queue(JobKind.COPY)
        await self.check_threadpool_executors_is_empty(self.job_executor_consumer._export_threadpool_executors)

    async def test_branching_workers_shutdown(self):
        await self.check_discover_new_queue(JobKind.REGRESSION)
        await self.check_threadpool_executors_is_empty(self.job_executor_consumer._branching_threadpool_executors)

    async def test_discover_new_import_query_and_export_queue(self):
        job1 = FakeTestJob(JobKind.IMPORT, self.user)
        job2 = FakeTestJob(JobKind.DELETE_DATA, self.user)
        job3 = FakeTestJob(JobKind.COPY, self.user)
        job4 = FakeTestJob(JobKind.REGRESSION, self.user)
        self.job_executor_producer.put_job(job1)
        self.job_executor_producer.put_job(job2)
        self.job_executor_producer.put_job(job3)
        self.job_executor_producer.put_job(job4)
        self.assertTrue(self.user.database_server in self.job_executor_producer._import_threadpool_executors)
        self.assertTrue(self.user.database_server in self.job_executor_producer._query_threadpool_executors)
        self.assertTrue(self.user.database_server in self.job_executor_producer._export_threadpool_executors)
        self.assertTrue(self.user.database_server in self.job_executor_producer._branching_threadpool_executors)
        self.assertEqual(self.job_executor_producer._import_threadpool_executors[self.user.database_server]._thread_pool_executor, None)
        self.assertEqual(self.job_executor_producer._query_threadpool_executors[self.user.database_server]._thread_pool_executor, None)
        self.assertEqual(self.job_executor_producer._export_threadpool_executors[self.user.database_server]._thread_pool_executor, None)
        self.assertEqual(self.job_executor_producer._branching_threadpool_executors[self.user.database_server]._thread_pool_executor, None)
        await self.check_threadpool_exists(self.job_executor_consumer._import_threadpool_executors)
        await self.check_threadpool_exists(self.job_executor_consumer._query_threadpool_executors)
        await self.check_threadpool_exists(self.job_executor_consumer._export_threadpool_executors)
        await self.check_threadpool_exists(self.job_executor_consumer._branching_threadpool_executors)

    async def test_all_workers_shutdown(self):
        await self.test_discover_new_import_query_and_export_queue()
        await self.check_threadpool_executors_is_empty(self.job_executor_consumer._import_threadpool_executors)
        await self.check_threadpool_executors_is_empty(self.job_executor_consumer._query_threadpool_executors)
        await self.check_threadpool_executors_is_empty(self.job_executor_consumer._export_threadpool_executors)
        await self.check_threadpool_executors_is_empty(self.job_executor_consumer._branching_threadpool_executors)

    async def test_import_workers_shutdown_after_job_finishes(self):
        job = FakeTestJob(JobKind.IMPORT, self.user, JobExecutor.ALIVE_QUEUES_TTL_IN_SECONDS + PYTHON_THREAD_TOLERANCE_IN_SECONDS)
        self.job_executor_producer.put_job(job)
        await self.check_threadpool_exists(self.job_executor_consumer._import_threadpool_executors)
        await asyncio.sleep(JobExecutor.ALIVE_QUEUES_TTL_IN_SECONDS)
        self.assertTrue(self.user.database_server in self.job_executor_consumer._import_threadpool_executors)
        await self.check_threadpool_executors_is_empty(self.job_executor_consumer._import_threadpool_executors)

    async def test_query_workers_shutdown_after_job_finishes(self):
        job = FakeTestJob(JobKind.QUERY, self.user, JobExecutor.ALIVE_QUEUES_TTL_IN_SECONDS + PYTHON_THREAD_TOLERANCE_IN_SECONDS)
        self.job_executor_producer.put_job(job)
        await self.check_threadpool_exists(self.job_executor_consumer._query_threadpool_executors)
        await asyncio.sleep(JobExecutor.ALIVE_QUEUES_TTL_IN_SECONDS)
        self.assertTrue(self.user.database_server in self.job_executor_consumer._query_threadpool_executors)
        await self.check_threadpool_executors_is_empty(self.job_executor_consumer._query_threadpool_executors)

    async def test_export_workers_shutdown_after_job_finishes(self):
        job = FakeTestJob(JobKind.COPY, self.user, JobExecutor.ALIVE_QUEUES_TTL_IN_SECONDS + PYTHON_THREAD_TOLERANCE_IN_SECONDS)
        self.job_executor_producer.put_job(job)
        await self.check_threadpool_exists(self.job_executor_consumer._export_threadpool_executors)
        await asyncio.sleep(JobExecutor.ALIVE_QUEUES_TTL_IN_SECONDS)
        self.assertTrue(self.user.database_server in self.job_executor_consumer._export_threadpool_executors)
        await self.check_threadpool_executors_is_empty(self.job_executor_consumer._export_threadpool_executors)

    async def test_branching_workers_shutdown_after_job_finishes(self):
        job = FakeTestJob(JobKind.REGRESSION, self.user, JobExecutor.ALIVE_QUEUES_TTL_IN_SECONDS + PYTHON_THREAD_TOLERANCE_IN_SECONDS)
        self.job_executor_producer.put_job(job)
        await self.check_threadpool_exists(self.job_executor_consumer._branching_threadpool_executors)
        await asyncio.sleep(JobExecutor.ALIVE_QUEUES_TTL_IN_SECONDS)
        self.assertTrue(self.user.database_server in self.job_executor_consumer._branching_threadpool_executors)
        await self.check_threadpool_executors_is_empty(self.job_executor_consumer._branching_threadpool_executors)

    async def test_long_import_worker_and_short_one_after_that(self):
        job1 = FakeTestJob(JobKind.IMPORT, self.user, JobExecutor.ALIVE_QUEUES_TTL_IN_SECONDS)
        job2 = FakeTestJob(JobKind.IMPORT, self.user)
        self.job_executor_producer.put_job(job1)
        self.job_executor_producer.put_job(job2)
        self.assertTrue(self.user.database_server in self.job_executor_producer._import_threadpool_executors)
        self.assertEqual(len(self.job_executor_producer._query_threadpool_executors), 0)
        self.assertEqual(self.job_executor_producer._import_threadpool_executors[self.user.database_server]._thread_pool_executor, None)
        await self.check_threadpool_exists(self.job_executor_consumer._import_threadpool_executors)
        await get_finalised_job_async(job1.id)
        await get_finalised_job_async(job2.id)
        await self.check_threadpool_exists(self.job_executor_consumer._import_threadpool_executors)
        await self.check_threadpool_executors_is_empty(self.job_executor_consumer._query_threadpool_executors)




import os.path
import unittest

from tinybird.csv_guess import guess_delimiter, guess_columns, has_header
from tinybird.csv_processing_queue import MAX_GUESS_BYTES, cut_csv_extract
from .utils import fixture_data


class TestCSVGuessDelimiter(unittest.TestCase):

    def test_csv_delimiter(self):
        data = fixture_data('movies.csv')
        delimiter = guess_delimiter(data)
        self.assertEqual(delimiter, ';')

    def test_csv_fact_venta_delimiter(self):
        data = fixture_data('fact_venta.csv')
        delimiter = guess_delimiter(data)
        self.assertEqual(delimiter, ',')

    def test_another_csv_delimiter(self):
        data = fixture_data('salas.csv')
        delimiter = guess_delimiter(data)
        self.assertEqual(delimiter, ';')

    def test_guess_float_column(self):
        data = fixture_data('float_column.csv')
        column_types = guess_columns(data, ',')

        guessed_types = [x['type'] for x in column_types]
        self.assertEqual(['Int64', 'Int64', 'Int64', 'Float32'], guessed_types)

    def test_guess_float_column_taxi(self):
        data = cut_csv_extract(fixture_data('float_column_taxi.csv')[:MAX_GUESS_BYTES], MAX_GUESS_BYTES)
        column_types = guess_columns(data, ',')

        guessed_types = [x['type'] for x in column_types]
        # last column is detected as String because they are all nulls
        actual_types = ['Int16', 'DateTime', 'DateTime', 'Int16', 'Float32', 'Int16', 'String', 'Int32', 'Int32', 'Int16', 'Float32', 'Float32', 'Float32', 'Float32', 'Float32', 'Float32', 'Float32', 'String']
        self.assertEqual(actual_types, guessed_types)

    def test_guess_float_column_with_string_as_string(self):
        data = fixture_data('float_column_with_string.csv')
        column_types = guess_columns(data, ',')

        guessed_types = [x['type'] for x in column_types]
        self.assertEqual(['Int64', 'Int64', 'Int64', 'String'], guessed_types)

    def test_has_header(self):
        with_headers = [
            "messy_headers.csv",
            "yt_1000.csv",
            "2918c.csv",
            "spanish_dates_0.csv",
            "small.csv",
            "yt_100.csv",
            "yt_1000_table_in_column.csv",
            "salas.csv",
            "parsing_error.csv",
            "bad_header.csv",
            "trans.csv",
            "shopify_export.csv",
            "CEND_14_13.csv",
            "float_column_with_string.csv",
            "float_column_taxi.csv",
            "tweet_activity_metrics_javisantana_20190220_20190320_en.csv",
            "arrays.csv",
            "movies.csv",
            "dates_with_decimals.csv",
            "sales_0.csv",
            "float_column.csv",
            "sales_1.csv",
            "sales_3.csv",
            "sales_2.csv",
        ]
        for file in with_headers:
            print(f"file with headers: {file}")
            data = fixture_data(os.path.basename(file))
            hh, _, _ = has_header(data, guess_delimiter(data))
            self.assertTrue(hh)

    def test_stop_iteration_exception(self):
        data = fixture_data('too_many_blank_lines.csv')
        column_types = guess_columns(data, ',')

        guessed_types = [x['type'] for x in column_types]
        expected_guessed_types = ['DateTime64', 'String', 'String']

        self.assertEqual(expected_guessed_types, guessed_types)

    def test_no_has_header(self):
        with_no_headers = [
            ("issue-920.csv", ','),
            ("issue-920-b.csv", ','),
        ]
        for (file, expected_delimiter) in with_no_headers:
            data = fixture_data(os.path.basename(file))
            delimiter = guess_delimiter(data)
            self.assertEqual(delimiter, expected_delimiter)
            hh, _, _ = has_header(data, delimiter)
            self.assertFalse(hh)


if __name__ == '__main__':
    unittest.main()



from datetime import datetime
from io import StringIO
from unittest.mock import patch

import pytest
import tornado
import uuid

from tests.utils import exec_sql
from tests.views.base_test import BaseTest, TBApiProxy, TBApiProxyAsync
from tinybird.job import WorkspaceDatabaseUsageTracker, UsageMetricsTracker
from tinybird.user import public, User, UserAccount, Users
from tinybird.user_workspace import UserWorkspaceRelationship
from tinybird.constants import Relationships
from tinybird.token_scope import scopes
from tinybird.ch import ch_create_table_as_table

from tests.conftest import is_main_process


@pytest.mark.serial  # Direct access to the public user
@pytest.mark.skipif(not is_main_process(), reason="Serial test")
class TestDataTracker(BaseTest):

    def setUp(self):
        super(TestDataTracker, self).setUp()
        self.original_user_get_all = User.get_all
        self.original_user_account_get_all = UserAccount.get_all
        self.original_relationships_get_all = UserWorkspaceRelationship.get_all

        self.tb_api_proxy = TBApiProxy(self)
        public_user = public.get_public_user()
        token = public_user.get_token_for_scope(scopes.ADMIN)
        self.tb_api_proxy.truncate_datasource(token, 'user_accounts_all')
        self.tb_api_proxy.truncate_datasource(token, 'workspaces_all')
        self.tb_api_proxy.truncate_datasource(token, 'user_workspaces_all')

    def tearDown(self):
        User.get_all = self.original_user_get_all
        UserAccount.get_all = self.original_user_account_get_all
        UserWorkspaceRelationship.get_all = self.original_relationships_get_all
        super(TestDataTracker, self).tearDown()

    def _execute_tracker_mocking_get_alls(self, user_get_all=None, user_account_get_all=None, relationships_get_all=None):

        if user_get_all:
            User.get_all = user_get_all

        if user_account_get_all:
            UserAccount.get_all = user_account_get_all

        if relationships_get_all:
            UserWorkspaceRelationship.get_all = relationships_get_all

        data_tracker = WorkspaceDatabaseUsageTracker()
        data_tracker.start()
        data_tracker.terminate()
        data_tracker.join()

    def test_data_tracker_updates_workspace_info(self):
        pu = public.get_public_user()
        workspaces_all_ds = pu.get_datasource('workspaces_all')

        created_at = datetime.now()
        self._execute_tracker_mocking_get_alls(user_get_all=lambda: [
            User(id='1234', name="1234_name", database="1234_database", database_server="1234_database_server",
                 created_at=created_at)
        ])

        content = exec_sql(pu.database, f"select * from {workspaces_all_ds.id} FORMAT JSON")
        self.assertEqual(content['data'], [{'id': '1234', 'name': '1234_name', 'database': '1234_database', 'database_server': '1234_database_server', 'plan': 'dev', 'deleted': 0, 'created_at': created_at.strftime('%Y-%m-%d %H:%M:%S')}])

        created_at = datetime.now()
        self._execute_tracker_mocking_get_alls(user_get_all=lambda: [
            User(id='5678', name="5678_name", database="5678_database", database_server="5678_database_server",
                 created_at=created_at)
        ])

        content = exec_sql(pu.database, f"select * from {workspaces_all_ds.id} FORMAT JSON")

        self.assertEqual(len(content['data']), 2)

        # Check ReplacingMergeTree merges correctly
        created_at = datetime.now()
        self._execute_tracker_mocking_get_alls(user_get_all=lambda: [
            User(id='5678', name="5678_name", database="5678_database", database_server="5678_database_server",
                 created_at=created_at)
        ])
        self.assertEqual(len(content['data']), 2)

    def test_data_tracker_correctly_writes_user_accounts(self):
        pu = public.get_public_user()
        user_accounts_ds = pu.get_datasource('user_accounts_all')

        created_at = datetime.now()
        self._execute_tracker_mocking_get_alls(user_account_get_all=lambda: [
            UserAccount(id='useracc1', email="useracc1_email", created_at=created_at)
        ])

        content = exec_sql(pu.database, f"select * from {user_accounts_ds.id} FORMAT JSON")
        self.assertEqual(content['data'], [{'id': 'useracc1', 'email': 'useracc1_email', 'feature_flags': '{}', 'deleted': 0, 'created_at': created_at.strftime('%Y-%m-%d %H:%M:%S')}])

    def test_data_tracker_correctly_writes_user_workspaces(self):
        pu = public.get_public_user()
        user_workspaces_ds = pu.get_datasource('user_workspaces_all')

        created_at = datetime.now()
        self._execute_tracker_mocking_get_alls(relationships_get_all=lambda: [
            UserWorkspaceRelationship(id="rel_id", user_id='rel_user_id', workspace_id="rel_works_id", relationship=Relationships.ADMIN, created_at=created_at)
        ])

        content = exec_sql(pu.database, f"select * from {user_workspaces_ds.id} FORMAT JSON")
        self.assertEqual(content['data'], [{'id': 'rel_id', 'relationship': 'admin', 'user_id': 'rel_user_id', 'workspace_id': 'rel_works_id', 'created_at': created_at.strftime('%Y-%m-%d %H:%M:%S')}])

    def test_data_tracker_correctly_writes_database_usage(self):
        pu = public.get_public_user()
        db_usage_ds = pu.get_datasource('db_usage')

        self._execute_tracker_mocking_get_alls()

        content = exec_sql(pu.database, f"select count(*) as total from {db_usage_ds.id} FORMAT JSON")
        self.assertGreater(int(content['data'][0]['total']), 0)


class TestUsageMetricsTracker(BaseTest):

    def setUp(self):
        super(TestUsageMetricsTracker, self).setUp()
        self.test_workspace = User.get_by_id(self.WORKSPACE_ID)
        self.test_user = UserAccount.get_by_id(self.USER_ID)
        self.tb_api_proxy_async = TBApiProxyAsync(self)

    @tornado.testing.gen_test
    @pytest.mark.skip(
        "usage_metrics_storage is going to be deprecated with https://gitlab.com/tinybird/analytics/-/issues/5280.")
    async def test_track_usage_metrics_storage(self):
        usage_metrics_tracker = UsageMetricsTracker()
        public_user = public.get_public_user()
        usage_metrics_storage = public_user.get_datasource('usage_metrics_storage')

        workspace_name = f"test_track_usage_{uuid.uuid4().hex}"
        email = f'{workspace_name}@example.com'
        workspace = await self.tb_api_proxy_async.register_user_and_workspace(email, workspace_name)
        self.workspaces_to_delete.append(workspace)
        token = Users.get_token_for_scope(workspace, scopes.ADMIN)

        await self.tb_api_proxy_async.create_datasource_from_data(
            token=token,
            ds_name='datasource_a',
            data=StringIO("d,sales\n2019-01-01,2\n2019-01-02,3")
        )
        await self.tb_api_proxy_async.create_datasource_from_data(
            token=token,
            ds_name='datasource_b',
            data=StringIO("d,sales\n2019-01-01,2\n2019-01-02,3")
        )

        datasource_a = Users.get_datasource(workspace, 'datasource_a')
        datasource_b = Users.get_datasource(workspace, 'datasource_b')
        self.wait_for_datasource_replication(workspace, datasource_a)
        self.wait_for_datasource_replication(workspace, datasource_b)

        usage_metrics_tracker.track_usage_metrics_storage(
            public_user,
            [User.get_by_id(workspace.id)]
        )

        self.wait_for_public_table_replication(usage_metrics_storage.name)
        content = exec_sql(public_user.database, f"select * from {usage_metrics_storage.id} FINAL WHERE database='{workspace.database}' FORMAT JSON")
        self.assertTrue(content['data'][0]['date'])
        self.assertTrue(content['data'][0]['database'] == workspace.database)
        self.assertEqual(int(content['data'][0]['rows']), 4)
        self.assertTrue(int(content['data'][0]['bytes_on_disk']) == pytest.approx(318, 50))

    @tornado.testing.gen_test
    async def test_track_usage_metrics_storage_v2(self):
        usage_metrics_tracker = UsageMetricsTracker()
        public_user = public.get_public_user()
        usage_metrics_storage = public_user.get_datasource('usage_metrics_storage__v2')

        workspace_name = f"test_track_usage_v2_{uuid.uuid4().hex}"
        email = f'{workspace_name}@example.com'
        workspace = await self.tb_api_proxy_async.register_user_and_workspace(email, workspace_name)
        user_account = UserAccount.get_by_email(email)
        token = Users.get_token_for_scope(workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(workspace)

        await self.tb_api_proxy_async.create_datasource_from_data(
            token=token,
            ds_name='datasource_a',
            data=StringIO("d,sales\n2019-01-01,2\n2019-01-02,3")
        )
        ds_a = Users.get_datasource(workspace, 'datasource_a')
        await self.tb_api_proxy_async.create_datasource_from_data(
            token=token,
            ds_name='datasource_b',
            data=StringIO("d,sales\n2019-01-01,2\n2019-01-02,3")
        )
        ds_b = Users.get_datasource(workspace, 'datasource_b')

        self.wait_for_datasource_replication(workspace, ds_a)
        self.wait_for_datasource_replication(workspace, ds_b)

        # create and share to test_workspace a ds
        ws_to_share_name = f'workspace_to_share_{uuid.uuid4().hex}'
        workspace_to_share = await self.tb_api_proxy_async.register_user_and_workspace(f'{ws_to_share_name}@example.com', ws_to_share_name)
        user_to_share = UserAccount.get_by_email(f'{ws_to_share_name}@example.com')
        token_workspace_to_share = Users.get_token_for_scope(workspace_to_share, scopes.ADMIN_USER)
        token_user_to_share = UserAccount.get_token_for_scope(user_to_share, scopes.AUTH)
        self.workspaces_to_delete.append(workspace_to_share)

        token_workspace_to_share = Users.get_token_for_scope(workspace_to_share, scopes.ADMIN)

        datasource_to_share_in_workspace_to_share = await self.tb_api_proxy_async.create_datasource(
            token=token_workspace_to_share,
            ds_name='datasource_to_share',
            schema='col_a Int32,col_b Int32,col_c Int32'
        )

        await self.tb_api_proxy_async.invite_user_to_workspace(
            token=UserAccount.get_token_for_scope(user_account, scopes.AUTH),
            workspace_id=workspace.id,
            user_to_invite_email=user_to_share.email
        )

        await self.tb_api_proxy_async.share_datasource_with_another_workspace(
            token=token_user_to_share,
            datasource_id=datasource_to_share_in_workspace_to_share['datasource']['id'],
            origin_workspace_id=workspace_to_share.id,
            destination_workspace_id=workspace.id
        )

        usage_metrics_tracker.track_usage_metrics_storage(
            public_user,
            [User.get_by_id(workspace.id)]
        )

        self.wait_for_public_table_replication(usage_metrics_storage.name)
        content = exec_sql(public_user.database, f"select * from {usage_metrics_storage.id} FINAL WHERE user_id='{workspace.id}' FORMAT JSON")
        self.assertEqual(len(content['data']), 2)

        for row in content['data']:
            self.assertIn(row['datasource_id'], [ds_b.id, ds_a.id])
            self.assertTrue(row['timestamp'])
            self.assertTrue(row['user_id'] == workspace.id)
            self.assertEqual(int(row['rows']), 2)
            self.assertEqual(int(row['bytes']), pytest.approx(155, 25))
            self.assertEqual(int(row['rows_quarantine']), 0)
            self.assertEqual(int(row['bytes_quarantine']), 0)

    @tornado.testing.gen_test
    @pytest.mark.skip(
        "usage_metrics_storage is going to be deprecated with https://gitlab.com/tinybird/analytics/-/issues/5280.")
    async def test_track_usage_metrics_storage_counts_quarantine_rows(self):
        usage_metrics_tracker = UsageMetricsTracker()
        public_user = public.get_public_user()
        usage_metrics_storage = public_user.get_datasource('usage_metrics_storage')

        workspace_name = f"test_track_usage_quarantine_{uuid.uuid4().hex}"
        email = f'{workspace_name}@example.com'
        workspace = await self.tb_api_proxy_async.register_user_and_workspace(email, workspace_name)
        token = Users.get_token_for_scope(workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(workspace)

        await self.tb_api_proxy_async.create_datasource_from_data(
            token=token,
            ds_name='datasource_a',
            data=StringIO("d,sales\n2019-01-01,2\n2019-01-02,3")
        )

        await self.tb_api_proxy_async.create_datasource_from_data(
            token=token,
            ds_name='datasource_a',
            data=StringIO("d,sales\nwrong_date,2\nwrong_date,3"),
            extra_params={'mode': 'append'}
        )
        ds_a = Users.get_datasource(workspace, 'datasource_a')

        self.wait_for_datasource_replication(workspace, ds_a, quarantine=True)

        usage_metrics_tracker.track_usage_metrics_storage(
            public_user,
            [User.get_by_id(workspace.id)]
        )

        self.wait_for_public_table_replication(usage_metrics_storage.name)
        content = exec_sql(public_user.database, f"select * from {usage_metrics_storage.id} FINAL WHERE database='{workspace.database}' FORMAT JSON")

        self.assertTrue(content['data'][0]['date'])
        self.assertTrue(content['data'][0]['database'] == workspace.database)
        self.assertEqual(int(content['data'][0]['rows']), 4)
        self.assertTrue(int(content['data'][0]['bytes_on_disk']) == pytest.approx(704, 50))

    @tornado.testing.gen_test
    async def test_track_usage_metrics_storage_counts_quarantine_rows_v2(self):
        usage_metrics_tracker = UsageMetricsTracker()
        public_user = public.get_public_user()
        usage_metrics_storage = public_user.get_datasource('usage_metrics_storage__v2')

        workspace_name = f"test_track_usage_quarantine_{uuid.uuid4().hex}"
        email = f'{workspace_name}@example.com'
        workspace = await self.tb_api_proxy_async.register_user_and_workspace(email, workspace_name)
        token = Users.get_token_for_scope(workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(workspace)

        await self.tb_api_proxy_async.create_datasource_from_data(
            token=token,
            ds_name='datasource_a',
            data=StringIO("d,sales\n2019-01-01,2\n2019-01-02,3")
        )
        ds_a = Users.get_datasource(workspace, 'datasource_a')

        await self.tb_api_proxy_async.create_datasource_from_data(
            token=token,
            ds_name='datasource_a',
            data=StringIO("d,sales\nwrong_date,2\nwrong_date,3"),
            extra_params={'mode': 'append'}
        )

        ds_a = Users.get_datasource(workspace, 'datasource_a')

        self.wait_for_datasource_replication(workspace, ds_a, quarantine=True)

        usage_metrics_tracker.track_usage_metrics_storage(
            public_user,
            [User.get_by_id(workspace.id)]
        )

        self.wait_for_public_table_replication(usage_metrics_storage.name)
        content = exec_sql(public_user.database, f"select * from {usage_metrics_storage.id} FINAL WHERE user_id='{workspace.id}' FORMAT JSON")

        self.assertEqual(len(content['data']), 1)

        for row in content['data']:
            self.assertEqual(row['datasource_id'], ds_a.id)
            self.assertTrue(row['timestamp'])
            self.assertTrue(row['user_id'] == workspace.id)
            self.assertEqual(int(row['rows']), 2)
            self.assertEqual(int(row['bytes']), pytest.approx(155, 25))
            self.assertEqual(int(row['rows_quarantine']), 2)
            self.assertEqual(int(row['bytes_quarantine']), pytest.approx(155, 25))

    @tornado.testing.gen_test
    @pytest.mark.skip(
        "usage_metrics_storage is going to be deprecated with https://gitlab.com/tinybird/analytics/-/issues/5280.")
    async def test_track_usage_metrics_storage_filters_temporary_datasources(self):
        usage_metrics_tracker = UsageMetricsTracker()
        public_user = public.get_public_user()
        usage_metrics_storage = public_user.get_datasource('usage_metrics_storage')

        workspace_name = f"test_track_usage_quarantine_{uuid.uuid4().hex}"
        email = f'{workspace_name}@example.com'
        workspace = await self.tb_api_proxy_async.register_user_and_workspace(email, workspace_name)
        token = Users.get_token_for_scope(workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(workspace)

        response = await self.tb_api_proxy_async.create_datasource_from_data(
            token=token,
            ds_name='datasource_a',
            data=StringIO("d,sales\n2019-01-01,2\n2019-01-02,3")
        )

        datasource_a = response['datasource']
        await ch_create_table_as_table(
            database_server=workspace.database_server,
            database=workspace.database,
            table_name='t_1234_tmp_1234',
            as_table_name=datasource_a['id'],
            storage_policy=workspace.storage_policy,
            engine='Null()',
            cluster='tinybird',
            as_table_database=workspace.database)
        exec_sql(workspace.database, f"INSERT INTO {workspace.database}.t_1234_tmp_1234 VALUES ('2019-01-01',4)")
        await ch_create_table_as_table(
            database_server=workspace.database_server,
            database=workspace.database,
            table_name='t_1234_mirror',
            as_table_name=datasource_a['id'],
            storage_policy=workspace.storage_policy,
            engine='Null()',
            cluster='tinybird',
            as_table_database=workspace.database)
        exec_sql(workspace.database, f"INSERT INTO {workspace.database}.t_1234_mirror VALUES ('2019-01-01',4)")

        usage_metrics_tracker.track_usage_metrics_storage(
            public_user,
            [User.get_by_id(workspace.id)]
        )

        self.wait_for_public_table_replication(usage_metrics_storage.name)
        content = exec_sql(public_user.database, f"select * from {usage_metrics_storage.id} FINAL WHERE database='{workspace.database}' FORMAT JSON")

        self.assertTrue(content['data'][0]['date'])
        self.assertTrue(content['data'][0]['database'] == workspace.database)
        self.assertTrue(int(content['data'][0]['rows']) == 2)
        self.assertTrue(int(content['data'][0]['bytes_on_disk']) == pytest.approx(159, 50))

    @tornado.testing.gen_test
    @patch.object(UsageMetricsTracker, '_retrieve_storage_used_per_database_and_table')
    async def test_track_usage_metrics_storage_no_cluster_assigned_not_found(self, mock_retrieve):
        usage_metrics_tracker = UsageMetricsTracker()
        public_user = public.get_public_user()
        usage_metrics_storage = public_user.get_datasource('usage_metrics_storage__v2')

        workspace_name = f"test_track_usage_quarantine_{uuid.uuid4().hex}"
        email = f'{workspace_name}@example.com'
        workspace = await self.tb_api_proxy_async.register_user_and_workspace(email, workspace_name)
        token = Users.get_token_for_scope(workspace, scopes.ADMIN)
        self.workspaces_to_delete.append(workspace)

        await self.tb_api_proxy_async.create_datasource_from_data(
            token=token,
            ds_name='datasource_a',
            data=StringIO("d,sales\n2019-01-01,2\n2019-01-02,3")
        )
        ds_a = Users.get_datasource(workspace, 'datasource_a')

        self.wait_for_datasource_replication(workspace, ds_a)

        mock_retrieve.return_value = {'cluster1': {'other_db': {'other_ds': {'rows': 1, 'bytes_on_disk': 10},
                                                                "other_ds_quarantine": {'rows': 2, 'bytes_on_disk': 20}}}}

        usage_metrics_tracker.track_usage_metrics_storage(
            public_user,
            [User.get_by_id(workspace.id)]
        )
        self.wait_for_public_table_replication(usage_metrics_storage.name)
        content = exec_sql(public_user.database, f"select * from {usage_metrics_storage.id} FINAL WHERE user_id='{workspace.id}' FORMAT JSON")
        self.assertEqual(len(content['data']), 0)



from contextvars import ContextVar


workspace_id: ContextVar[str] = ContextVar('workspace_id')
workspace: ContextVar[str] = ContextVar('workspace')
table_id: ContextVar[str] = ContextVar('table_id')
hfi_frequency: ContextVar[float] = ContextVar('hfi_frequency')
disable_template_security_validation: ContextVar[bool] = ContextVar('disable_template_security_validation')
origin: ContextVar[str] = ContextVar('origin')
request_id: ContextVar[str] = ContextVar('request_id')



import unittest

from tinybird.feature_flags import FeatureFlagsService, FeatureFlagBase, FeatureDetails, FeatureFlagsWorkspaceService


class FeatureFlagMock(FeatureFlagBase):
    FULL_ACCESS_ACCOUNT = "full_access_account"
    ONBOARDING = "onboarding"


class TestFeatureFlags(unittest.TestCase):

    _configured_domain = None
    _configured_map_features_and_details = None

    @classmethod
    def setUpClass(cls) -> None:

        cls._configured_domain = FeatureFlagsService.configured_domain
        cls._configured_map_features_and_details = FeatureFlagsService.map_features_and_details

        FeatureFlagsService.configured_domain = 'tinybird.co'

        FeatureFlagsService.map_features_and_details = {
            FeatureFlagMock.FULL_ACCESS_ACCOUNT: FeatureDetails(
                "Feature default false but true for configured domain, hidden", False, override_for_configured_domain=True),
            FeatureFlagMock.ONBOARDING: FeatureDetails(
                "not hidden one", False, private=False, override_for_configured_domain=True),
        }

    @classmethod
    def tearDownClass(cls) -> None:
        FeatureFlagsService.configured_domain = cls._configured_domain
        FeatureFlagsService.map_features_and_details = cls._configured_map_features_and_details
        super().tearDownClass()

    def test_get_existing_feature_for_a_not_configured_domain(self):
        ff_value = FeatureFlagsService.feature_for_email(FeatureFlagMock.FULL_ACCESS_ACCOUNT, 'random@email.com')
        self.assertEqual(False, ff_value)

    def test_get_existing_feature_for_a_not_configured_domain_but_with_config_override(self):
        ff_value = FeatureFlagsService.feature_for_email(FeatureFlagMock.FULL_ACCESS_ACCOUNT, 'random@email.com',
                                                         {FeatureFlagMock.FULL_ACCESS_ACCOUNT.value: True})
        self.assertEqual(True, ff_value)

    def test_get_existing_feature_for_a_configured_domain(self):
        ff_value = FeatureFlagsService.feature_for_email(FeatureFlagMock.FULL_ACCESS_ACCOUNT, 'random@tinybird.co')
        self.assertEqual(True, ff_value)

    def test_get_existing_feature_that_is_overrided(self):
        ff_value = FeatureFlagsService.feature_for_email(FeatureFlagMock.ONBOARDING, 'random@tinybird.co', {FeatureFlagMock.ONBOARDING.value: True})
        self.assertEqual(True, ff_value)

    def test_get_all_features_for_private_listing(self):
        ff_values = FeatureFlagsService.get_all_feature_values('random@email.com', include_private=True)

        self.assertEqual({
            'full_access_account': False,
            'onboarding': False,
        }, ff_values)

    def test_get_all_features_for_public_listing(self):
        ff_values = FeatureFlagsService.get_all_feature_values('random@email.com')

        self.assertEqual({
            'onboarding': False
        }, ff_values)


class FeatureFlagWorkspaceMock(FeatureFlagBase):
    DATASOURCES_OPS_LOG_ENRICHMENT = 'datasources_ops_log_enrichment'
    ENABLE_STORAGE_POLICY = 'enable_storage_policy'


class TestFeatureFlagsWorkspace(unittest.TestCase):

    _configured_map_features_and_details = None

    @classmethod
    def setUpClass(cls) -> None:

        cls._configured_map_features_and_details = FeatureFlagsWorkspaceService.map_features_and_details

        FeatureFlagsWorkspaceService.map_features_and_details = {
            FeatureFlagWorkspaceMock.DATASOURCES_OPS_LOG_ENRICHMENT: FeatureDetails(
                "Enable datasources ops log enrichment", default_value=False,
                private=False),
            FeatureFlagWorkspaceMock.ENABLE_STORAGE_POLICY: FeatureDetails(
                "Enable custom storage policy", default_value=False,
                private=True),
        }

    @classmethod
    def tearDownClass(cls) -> None:
        FeatureFlagsWorkspaceService.map_features_and_details = cls._configured_map_features_and_details
        super().tearDownClass()

    def test_get_existing_feature(self):
        ff_value = FeatureFlagsWorkspaceService.feature_for_id(FeatureFlagWorkspaceMock.DATASOURCES_OPS_LOG_ENRICHMENT, "", None)
        self.assertEqual(False, ff_value)

    def test_get_existing_feature_that_is_overrided(self):
        ff_value = FeatureFlagsWorkspaceService.feature_for_id(FeatureFlagWorkspaceMock.DATASOURCES_OPS_LOG_ENRICHMENT, '', {FeatureFlagWorkspaceMock.DATASOURCES_OPS_LOG_ENRICHMENT.value: True})
        self.assertEqual(True, ff_value)

    def test_get_all_features_for_private_listing(self):
        ff_values = FeatureFlagsWorkspaceService.get_all_feature_values(include_private=True)

        self.assertEqual({
            FeatureFlagWorkspaceMock.DATASOURCES_OPS_LOG_ENRICHMENT.value: False,
            FeatureFlagWorkspaceMock.ENABLE_STORAGE_POLICY.value: False,
        }, ff_values)

    def test_get_all_features_for_public_listing(self):
        ff_values = FeatureFlagsWorkspaceService.get_all_feature_values(include_private=False)

        self.assertEqual({
            FeatureFlagWorkspaceMock.DATASOURCES_OPS_LOG_ENRICHMENT.value: False
        }, ff_values)
